__all__ = ["StatisticalAnalyzer"]

from typing import *

from PySide6 import QtCore, QtGui, QtWidgets

from ..statistics import all_statistics
from ..models import Dataset, Sample
from ..charts import (highlight_color, FrequencyChart, Frequency3DChart, FrequencyHeatmap, CumulativeChart,
                      DiagramChart, BP12SSCDiagramChart, BP12GSMDiagramChart, Folk54SSCDiagramChart,
                      Folk54GSMDiagramChart, CMDiagramChart)


class StatisticalAnalyzer(QtWidgets.QWidget):
    PAGE_ROWS = 20

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._dataset: Optional[Dataset] = None
        self.data_table = QtWidgets.QTableWidget(100, 100)
        self.data_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.data_table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.data_table.setAlternatingRowColors(True)
        self.data_table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.data_table.setHorizontalHeaderLabels([self.tr("Tips")])
        for row, tip in enumerate(self.tips):
            item = QtWidgets.QTableWidgetItem(tip)
            item.setTextAlignment(QtCore.Qt.AlignCenter)
            self.data_table.setItem(row, 0, item)
        self.data_table.setColumnWidth(0, 800)
        self.data_table.setColumnCount(1)
        self.data_table.setRowCount(len(self.tips))
        self.main_layout = QtWidgets.QGridLayout(self)
        self.main_layout.addWidget(self.data_table, 0, 0, 1, 3)
        self.previous_button = QtWidgets.QPushButton(self.tr("Previous"))
        self.previous_button.setToolTip(self.tr("Click to back to the previous page."))
        self.previous_button.clicked.connect(self.on_previous_button_clicked)
        self.current_page_combo_box = QtWidgets.QComboBox()
        self.current_page_combo_box.addItem(self.tr("No Page"))
        _ = self.tr("Page {0}").format(1)
        self.current_page_combo_box.currentIndexChanged.connect(self.update_page)
        self.next_button = QtWidgets.QPushButton(self.tr("Next"))
        self.next_button.setToolTip(self.tr("Click to jump to the next page."))
        self.next_button.clicked.connect(self.on_next_button_clicked)
        self.main_layout.addWidget(self.previous_button, 1, 0)
        self.main_layout.addWidget(self.current_page_combo_box, 1, 1)
        self.main_layout.addWidget(self.next_button, 1, 2)
        self.geometric_checkbox = QtWidgets.QCheckBox(self.tr("Geometric [unit is {0}]").format("μm"))
        self.geometric_checkbox.setChecked(True)
        self.geometric_checkbox.stateChanged.connect(self.on_is_geometric_changed)
        self.main_layout.addWidget(self.geometric_checkbox, 2, 0)
        self.FW57_checkbox = QtWidgets.QCheckBox(self.tr("Method of Statistical Moments"))
        self.FW57_checkbox.setChecked(False)
        self.FW57_checkbox.stateChanged.connect(self.on_is_FW57_changed)
        self.main_layout.addWidget(self.FW57_checkbox, 2, 1)
        self.proportion_combo_box = QtWidgets.QComboBox()
        self.proportion_combo_box.addItems([description for _, description in self.supported_proportions])
        self.proportion_combo_box.currentIndexChanged.connect(lambda: self.update_page(self.page_index))
        self.main_layout.addWidget(self.proportion_combo_box, 2, 2)
        self.main_layout.setColumnStretch(0, 1)
        self.main_layout.setColumnStretch(1, 1)
        self.main_layout.setColumnStretch(2, 1)
        self.menu = QtWidgets.QMenu(self.data_table)
        self.plot_cumulative_menu = self.menu.addMenu(self.tr("Plot Cumulative Frequency Chart"))
        self.cumulative_plot_selected_action = self.plot_cumulative_menu.addAction(self.tr("Plot"))
        self.cumulative_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.cumulative_chart, self.selections, False))
        self.cumulative_append_selected_action = self.plot_cumulative_menu.addAction(self.tr("Append"))
        self.cumulative_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.cumulative_chart, self.selections, True))
        self.cumulative_plot_all_action = self.plot_cumulative_menu.addAction(self.tr("Plot All"))
        self.cumulative_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.cumulative_chart, self._dataset, False))
        self.cumulative_append_all_action = self.plot_cumulative_menu.addAction(self.tr("Append All"))
        self.cumulative_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.cumulative_chart, self._dataset, True))
        self.plot_frequency_menu = self.menu.addMenu(self.tr("Plot Frequency Distribution Chart"))
        self.frequency_plot_selected_action = self.plot_frequency_menu.addAction(self.tr("Plot"))
        self.frequency_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_chart, self.selections, False))
        self.frequency_append_selected_action = self.plot_frequency_menu.addAction(self.tr("Append"))
        self.frequency_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_chart, self.selections, True))
        self.frequency_plot_all_action = self.plot_frequency_menu.addAction(self.tr("Plot All"))
        self.frequency_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_chart, self._dataset, False))
        self.frequency_append_all_action = self.plot_frequency_menu.addAction(self.tr("Append All"))
        self.frequency_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_chart, self._dataset, True))
        self.plot_frequency_3D_menu = self.menu.addMenu(self.tr("Plot Frequency 3D Chart"))
        self.frequency_3D_plot_selected_action = self.plot_frequency_3D_menu.addAction(self.tr("Plot"))
        self.frequency_3D_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_3D_chart, self.selections, False))
        self.frequency_3D_append_selected_action = self.plot_frequency_3D_menu.addAction(self.tr("Append"))
        self.frequency_3D_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_3D_chart, self.selections, True))
        self.frequency_3D_plot_all_action = self.plot_frequency_3D_menu.addAction(self.tr("Plot All"))
        self.frequency_3D_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_3D_chart, self._dataset, False))
        self.frequency_3D_append_all_action = self.plot_frequency_3D_menu.addAction(self.tr("Append All"))
        self.frequency_3D_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_3D_chart, self._dataset, True))
        self.plot_frequency_heatmap_menu = self.menu.addMenu(self.tr("Plot Frequency Heatmap"))
        self.frequency_heatmap_plot_selected_action = self.plot_frequency_heatmap_menu.addAction(self.tr("Plot"))
        self.frequency_heatmap_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_heatmap, self.selections, False))
        self.frequency_heatmap_append_selected_action = self.plot_frequency_heatmap_menu.addAction(self.tr("Append"))
        self.frequency_heatmap_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_heatmap, self.selections, True))
        self.frequency_heatmap_plot_all_action = self.plot_frequency_heatmap_menu.addAction(self.tr("Plot All"))
        self.frequency_heatmap_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_heatmap, self._dataset, False))
        self.frequency_heatmap_append_all_action = self.plot_frequency_heatmap_menu.addAction(self.tr("Append All"))
        self.frequency_heatmap_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.frequency_heatmap, self._dataset, True))
        self.folk54_GSM_menu = self.menu.addMenu(self.tr("Plot GSM Diagram (Folk, 1954)"))
        self.folk54_GSM_plot_selected_action = self.folk54_GSM_menu.addAction(self.tr("Plot"))
        self.folk54_GSM_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.folk54_GSM_diagram, self.selections, False))
        self.folk54_GSM_append_selected_action = self.folk54_GSM_menu.addAction(self.tr("Append"))
        self.folk54_GSM_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.folk54_GSM_diagram, self.selections, True))
        self.folk54_GSM_plot_all_action = self.folk54_GSM_menu.addAction(self.tr("Plot All"))
        self.folk54_GSM_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.folk54_GSM_diagram, self._dataset, False))
        self.folk54_GSM_append_all_action = self.folk54_GSM_menu.addAction(self.tr("Append All"))
        self.folk54_GSM_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.folk54_GSM_diagram, self._dataset, True))
        self.folk54_SSC_menu = self.menu.addMenu(self.tr("Plot SSC Diagram (Folk, 1954)"))
        self.folk54_SSC_plot_selected_action = self.folk54_SSC_menu.addAction(self.tr("Plot"))
        self.folk54_SSC_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.folk54_SSC_diagram, self.selections, False))
        self.folk54_SSC_append_selected_action = self.folk54_SSC_menu.addAction(self.tr("Append"))
        self.folk54_SSC_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.folk54_SSC_diagram, self.selections, True))
        self.folk54_SSC_plot_all_action = self.folk54_SSC_menu.addAction(self.tr("Plot All"))
        self.folk54_SSC_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.folk54_SSC_diagram, self._dataset, False))
        self.folk54_SSC_append_all_action = self.folk54_SSC_menu.addAction(self.tr("Append All"))
        self.folk54_SSC_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.folk54_SSC_diagram, self._dataset, True))
        self.BP12_GSM_menu = self.menu.addMenu(self.tr("Plot GSM Diagram (Blott and Pye, 2012)"))
        self.BP12_GSM_plot_selected_action = self.BP12_GSM_menu.addAction(self.tr("Plot"))
        self.BP12_GSM_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.BP12_GSM_diagram, self.selections, False))
        self.BP12_GSM_append_selected_action = self.BP12_GSM_menu.addAction(self.tr("Append"))
        self.BP12_GSM_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.BP12_GSM_diagram, self.selections, True))
        self.BP12_GSM_plot_all_action = self.BP12_GSM_menu.addAction(self.tr("Plot All"))
        self.BP12_GSM_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.BP12_GSM_diagram, self._dataset, False))
        self.BP12_GSM_append_all_action = self.BP12_GSM_menu.addAction(self.tr("Append All"))
        self.BP12_GSM_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.BP12_GSM_diagram, self._dataset, True))
        self.BP12_SSC_menu = self.menu.addMenu(self.tr("Plot SSC Diagram (Blott and Pye, 2012)"))
        self.BP12_SSC_plot_selected_action = self.BP12_SSC_menu.addAction(self.tr("Plot"))
        self.BP12_SSC_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.BP12_SSC_diagram, self.selections, False))
        self.BP12_SSC_append_selected_action = self.BP12_SSC_menu.addAction(self.tr("Append"))
        self.BP12_SSC_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.BP12_SSC_diagram, self.selections, True))
        self.BP12_SSC_plot_all_action = self.BP12_SSC_menu.addAction(self.tr("Plot All"))
        self.BP12_SSC_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.BP12_SSC_diagram, self._dataset, False))
        self.BP12_SSC_append_all_action = self.BP12_SSC_menu.addAction(self.tr("Append All"))
        self.BP12_SSC_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.BP12_SSC_diagram, self._dataset, True))
        self.CM_menu = self.menu.addMenu(self.tr("Plot C-M Diagram"))
        self.CM_plot_selected_action = self.CM_menu.addAction(self.tr("Plot"))
        self.CM_plot_selected_action.triggered.connect(
            lambda: self.plot_chart(self.CM_diagram, self.selections, False))
        self.CM_append_selected_action = self.CM_menu.addAction(self.tr("Append"))
        self.CM_append_selected_action.triggered.connect(
            lambda: self.plot_chart(self.CM_diagram, self.selections, True))
        self.CM_plot_all_action = self.CM_menu.addAction(self.tr("Plot All"))
        self.CM_plot_all_action.triggered.connect(
            lambda: self.plot_chart(self.CM_diagram, self._dataset, False))
        self.CM_append_all_action = self.CM_menu.addAction(self.tr("Append All"))
        self.CM_append_all_action.triggered.connect(
            lambda: self.plot_chart(self.CM_diagram, self._dataset, True))
        self.previous_button.setEnabled(False)
        self.current_page_combo_box.setEnabled(False)
        self.next_button.setEnabled(False)
        self.data_table.customContextMenuRequested.connect(self.show_menu)
        self.cumulative_chart = CumulativeChart()
        self.frequency_chart = FrequencyChart()
        self.frequency_3D_chart = Frequency3DChart()
        self.frequency_heatmap = FrequencyHeatmap()
        self.folk54_GSM_diagram = Folk54GSMDiagramChart()
        self.folk54_SSC_diagram = Folk54SSCDiagramChart()
        self.BP12_GSM_diagram = BP12GSMDiagramChart()
        self.BP12_SSC_diagram = BP12SSCDiagramChart()
        self.CM_diagram = CMDiagramChart()
        self.normal_msg = QtWidgets.QMessageBox(self)

    def show_menu(self, pos):
        self.menu.popup(QtGui.QCursor.pos())

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_dataset_loaded(self, dataset: Dataset):
        if dataset is None:
            return
        self._dataset = dataset
        self.current_page_combo_box.clear()
        page_count, left = divmod(len(self._dataset), self.PAGE_ROWS)
        if left != 0:
            page_count += 1
        self.current_page_combo_box.addItems([self.tr("Page {0}").format(i + 1) for i in range(page_count)])
        self.previous_button.setEnabled(True)
        self.current_page_combo_box.setEnabled(True)
        self.next_button.setEnabled(True)

        self.update_page(0)

    @property
    def tips(self) -> List[str]:
        tips = [
            self.tr("By clicking the option at menu bar, you can load the grain size distributions \
                (Menu -> Open -> Grain Size Dataset)."),
            self.tr("By clicking the option at menu bar, you can save the statistical parameters and \
                classification groups to a Excel file (Menu -> Save -> Statistical Result)."),
            self.tr("By right clicking at the table region, you can open the menu to draw charts.")]
        return tips

    @property
    def is_geometric(self) -> bool:
        return self.geometric_checkbox.isChecked()

    def on_is_geometric_changed(self, state):
        if state == QtCore.Qt.Checked:
            self.geometric_checkbox.setText(self.tr("Geometric [unit is {0}]").format(self.unit))
        else:
            self.geometric_checkbox.setText(self.tr("Logarithmic [unit is {0}]").format(self.unit))
        self.update_page(self.page_index)

    @property
    def is_FW57(self) -> bool:
        return self.FW57_checkbox.isChecked()

    def on_is_FW57_changed(self, state):
        if state == QtCore.Qt.Checked:
            self.FW57_checkbox.setText(self.tr("Method of Folk and Ward (1957)"))
        else:
            self.FW57_checkbox.setText(self.tr("Method of Statistical Moments"))
        self.update_page(self.page_index)

    @property
    def supported_proportions(self) -> Sequence[Tuple[str, str]]:
        result = (
            ("proportions_gsm", self.tr("Gravel, Sand, Mud")),
            ("proportions_ssc", self.tr("Sand, Silt, Clay")),
            ("proportions_bgssc", self.tr("Boulder, Gravel, Sand, Silt, Clay")))
        return result

    @property
    def proportion(self) -> Tuple[str, str]:
        index = self.proportion_combo_box.currentIndex()
        key, description = self.supported_proportions[index]
        return key, description

    @property
    def page_index(self) -> int:
        return self.current_page_combo_box.currentIndex()

    @property
    def n_pages(self) -> int:
        return self.current_page_combo_box.count()

    @property
    def unit(self) -> str:
        return "μm" if self.is_geometric else "φ"

    def update_page(self, page_index: int):
        if self._dataset is None:
            return

        def write(row: int, col: int, value: Union[int, float, str]):
            if isinstance(value, str):
                pass
            elif isinstance(value, int):
                value = str(value)
            elif isinstance(value, float):
                value = f"{value: 0.2f}"
            else:
                value = value.__str__()
            item = QtWidgets.QTableWidgetItem(value)
            item.setTextAlignment(QtCore.Qt.AlignCenter)
            self.data_table.setItem(row, col, item)

        # necessary to clear
        self.data_table.clear()
        if page_index == self.n_pages - 1:
            start = page_index * self.PAGE_ROWS
            end = len(self._dataset)
        else:
            start, end = page_index * self.PAGE_ROWS, (page_index + 1) * self.PAGE_ROWS
        proportion_key, proportion_name = self.proportion
        col_names = [self.tr("Mean [{0}]").format(self.unit),
                     self.tr("Mean Description"),
                     self.tr("Median [{0}]").format(self.unit),
                     self.tr("Modes [{0}]").format(self.unit),
                     self.tr("Sorting Coefficient"),
                     self.tr("Sorting Description"),
                     self.tr("Skewness"),
                     self.tr("Skewness Description"),
                     self.tr("Kurtosis"),
                     self.tr("Kurtosis Description"),
                     self.tr("({0}) Proportion [%]").format(proportion_name),
                     self.tr("Group (Folk, 1954)"),
                     self.tr("Group Symbol (Blott and Pye, 2012)"),
                     self.tr("Group (Blott and Pye, 2012)")]
        col_keys = [(True, "mean"),
                    (True, "mean_description"),
                    (True, "median"),
                    (True, "modes"),
                    (True, "std"),
                    (True, "std_description"),
                    (True, "skewness"),
                    (True, "skewness_description"),
                    (True, "kurtosis"),
                    (True, "kurtosis_description"),
                    (False, proportion_key),
                    (False, "group_folk54"),
                    (False, "group_bp12_symbol"),
                    (False, "group_bp12")]
        self.data_table.setRowCount(end - start)
        self.data_table.setColumnCount(len(col_names))
        self.data_table.setHorizontalHeaderLabels(col_names)
        self.data_table.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.TextWordWrap)
        self.data_table.setVerticalHeaderLabels(self._dataset.sample_names[start: end])
        for row, sample in enumerate(self._dataset[start: end]):
            statistics = all_statistics(sample.classes, sample.classes_phi, sample.distribution)
            if self.is_geometric:
                if self.is_FW57:
                    sub_key = "geometric_fw57"
                else:
                    sub_key = "geometric"
            else:
                if self.is_FW57:
                    sub_key = "logarithmic_fw57"
                else:
                    sub_key = "logarithmic"
            for col, (in_sub, key) in enumerate(col_keys):
                value = statistics[sub_key][key] if in_sub else statistics[key]
                if key == "modes":
                    write(row, col, ", ".join([f"{m:0.2f}" for m in value]))
                elif key[:11] == "proportions":
                    write(row, col, ", ".join([f"{p * 100:0.2f}" for p in value]))
                else:
                    write(row, col, value)

        self.data_table.resizeColumnsToContents()

    @property
    def selections(self):
        if self._dataset is None:
            return []
        start = self.page_index * self.PAGE_ROWS
        temp = set()
        for item in self.data_table.selectedRanges():
            for i in range(item.topRow(), min(self.PAGE_ROWS + 1, item.bottomRow() + 1)):
                temp.add(i + start)
        indexes = list(temp)
        indexes.sort()
        samples = [self._dataset[i] for i in indexes]
        return samples

    def on_previous_button_clicked(self):
        if self.page_index > 0:
            self.current_page_combo_box.setCurrentIndex(self.page_index - 1)

    def on_next_button_clicked(self):
        if self.page_index < self.n_pages - 1:
            self.current_page_combo_box.setCurrentIndex(self.page_index + 1)

    def plot_chart(self, chart, samples: Iterable[Sample], append: bool):
        copy_samples = list(samples)
        if self._dataset is None:
            self.show_error(self.tr("The dataset has not been loaded."))
        elif len(copy_samples) == 0:
            self.show_error(self.tr("No sample was selected."))
        else:
            if isinstance(chart, DiagramChart):
                kwargs = {"mfc": highlight_color()}
            else:
                kwargs = {}
            chart.show_samples(copy_samples, append=append, **kwargs)
            chart.show()

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.data_table.setHorizontalHeaderLabels([self.tr("Tips")])
        if self._dataset is None:
            for row, tip in enumerate(self.tips):
                item = QtWidgets.QTableWidgetItem(tip)
                item.setTextAlignment(QtCore.Qt.AlignCenter)
                self.data_table.setItem(row, 0, item)
            self.current_page_combo_box.setItemText(0, self.tr("No Page"))
        else:
            self.update_page(self.page_index)
            for i in range(self.n_pages):
                self.current_page_combo_box.setItemText(i, self.tr("Page {0}").format(i + 1))
        self.previous_button.setText(self.tr("Previous"))
        self.previous_button.setToolTip(self.tr("Click to back to the previous page."))
        self.next_button.setText(self.tr("Next"))
        self.next_button.setToolTip(self.tr("Click to jump to the next page."))
        if self.is_geometric:
            self.geometric_checkbox.setText(self.tr("Geometric [unit is {0}]").format(self.unit))
        else:
            self.geometric_checkbox.setText(self.tr("Logarithmic [unit is {0}]").format(self.unit))
        if self.is_FW57:
            self.FW57_checkbox.setText(self.tr("Method of Folk and Ward (1957)"))
        else:
            self.FW57_checkbox.setText(self.tr("Method of Statistical Moments"))
        for i, (_, description) in enumerate(self.supported_proportions):
            self.proportion_combo_box.setItemText(i, description)
        self.plot_cumulative_menu.setTitle(self.tr("Plot Cumulative Frequency Chart"))
        self.cumulative_plot_selected_action.setText(self.tr("Plot"))
        self.cumulative_append_selected_action.setText(self.tr("Append"))
        self.cumulative_plot_all_action.setText(self.tr("Plot All"))
        self.cumulative_append_all_action.setText(self.tr("Append All"))
        self.plot_frequency_menu.setTitle(self.tr("Plot Frequency Distribution Chart"))
        self.frequency_plot_selected_action.setText(self.tr("Plot"))
        self.frequency_append_selected_action.setText(self.tr("Append"))
        self.frequency_plot_all_action.setText(self.tr("Plot All"))
        self.frequency_append_all_action.setText(self.tr("Append All"))
        self.plot_frequency_3D_menu.setTitle(self.tr("Plot Frequency 3D Chart"))
        self.frequency_3D_plot_selected_action.setText(self.tr("Plot"))
        self.frequency_3D_append_selected_action.setText(self.tr("Append"))
        self.frequency_3D_plot_all_action.setText(self.tr("Plot All"))
        self.frequency_3D_append_all_action.setText(self.tr("Append All"))
        self.plot_frequency_heatmap_menu.setTitle(self.tr("Plot Frequency Heatmap"))
        self.frequency_heatmap_plot_selected_action.setText(self.tr("Plot"))
        self.frequency_heatmap_append_selected_action.setText(self.tr("Append"))
        self.frequency_heatmap_plot_all_action.setText(self.tr("Plot All"))
        self.frequency_heatmap_append_all_action.setText(self.tr("Append All"))
        self.folk54_GSM_menu.setTitle(self.tr("Plot GSM Diagram (Folk, 1954)"))
        self.folk54_GSM_plot_selected_action.setText(self.tr("Plot"))
        self.folk54_GSM_append_selected_action.setText(self.tr("Append"))
        self.folk54_GSM_plot_all_action.setText(self.tr("Plot All"))
        self.folk54_GSM_append_all_action.setText(self.tr("Append All"))
        self.folk54_SSC_menu.setTitle(self.tr("Plot SSC Diagram (Folk, 1954)"))
        self.folk54_SSC_plot_selected_action.setText(self.tr("Plot"))
        self.folk54_SSC_append_selected_action.setText(self.tr("Append"))
        self.folk54_SSC_plot_all_action.setText(self.tr("Plot All"))
        self.folk54_SSC_append_all_action.setText(self.tr("Append All"))
        self.BP12_GSM_menu.setTitle(self.tr("Plot GSM Diagram (Blott and Pye, 2012)"))
        self.BP12_GSM_plot_selected_action.setText(self.tr("Plot"))
        self.BP12_GSM_append_selected_action.setText(self.tr("Append"))
        self.BP12_GSM_plot_all_action.setText(self.tr("Plot All"))
        self.BP12_GSM_append_all_action.setText(self.tr("Append All"))
        self.BP12_SSC_menu.setTitle(self.tr("Plot SSC Diagram (Blott and Pye, 2012)"))
        self.BP12_SSC_plot_selected_action.setText(self.tr("Plot"))
        self.BP12_SSC_append_selected_action.setText(self.tr("Append"))
        self.BP12_SSC_plot_all_action.setText(self.tr("Plot All"))
        self.BP12_SSC_append_all_action.setText(self.tr("Append All"))
        self.CM_menu.setTitle(self.tr("Plot C-M Diagram"))
        self.CM_plot_selected_action.setText(self.tr("Plot"))
        self.CM_append_selected_action.setText(self.tr("Append"))
        self.CM_plot_all_action.setText(self.tr("Plot All"))
        self.CM_append_all_action.setText(self.tr("Append All"))



import csv
import logging
import os
from enum import Enum, unique

import numpy as np
import openpyxl
import xlrd
from typing import *

from ..models.dataset import Dataset, validate_classes, validate_distributions


def check_layout(class_row: int, name_col: int, start_row: int, start_col: int):
    assert isinstance(class_row, int)
    assert isinstance(name_col, int)
    assert isinstance(start_row, int)
    assert isinstance(start_col, int)
    if class_row < 0 or \
            name_col < 0 or \
            start_row < 0 or \
            start_col < 0:
        raise ValueError("The index of row or column must be non-negative.")
    if class_row >= start_row:
        raise ValueError("The start row index of distributions must be greater than the row index of classes.")
    if name_col >= start_col:
        raise ValueError("The start column index of distributions must be greater "
                         "than the column index of sample names.")


@unique
class ReadFileType(Enum):
    XLS = 0
    XLSX = 1
    CSV = 2


def get_file_type(filename: str):
    _, extension = os.path.splitext(filename)
    if extension == ".csv":
        return ReadFileType.CSV
    elif extension == ".xls":
        return ReadFileType.XLS
    elif extension == ".xlsx":
        return ReadFileType.XLSX
    else:
        raise NotImplementedError(extension)


def _get_raw_table(
        filename: str, file_type: ReadFileType, sheet_index: int,
        progress_callback: Callable[[float], None] = None,
        logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)
    logger.debug(f"Try to load the raw data table from [{file_type.name}] file: {filename}.")
    if file_type == ReadFileType.CSV:
        with open(filename, encoding="utf-8") as f:
            reader = csv.reader(f)
            raw_table = []
            for i, row_values in enumerate(reader):
                raw_table.append(row_values)
                if progress_callback is not None:
                    progress_callback(i / reader.line_num)

    elif file_type == ReadFileType.XLS:
        workbook: xlrd.Book = xlrd.open_workbook(filename)
        sheet = workbook.sheet_by_index(sheet_index)
        raw_table = []
        for row in range(sheet.nrows):
            raw_table.append(sheet.row_values(row))
            if progress_callback is not None:
                progress_callback(row / sheet.nrows)

    elif file_type == ReadFileType.XLSX:
        workbook: openpyxl.Workbook = openpyxl.load_workbook(filename, read_only=True, data_only=True)
        sheet = workbook[workbook.sheetnames[sheet_index]]
        raw_table = []
        for i, row_values in enumerate(sheet.values):
            raw_table.append(row_values)
            if progress_callback is not None:
                progress_callback(i / sheet.max_row)
    else:
        raise NotImplementedError(file_type)
    logger.debug("The raw data table has been loaded from this file.")
    return raw_table


def load_dataset(filename: str,
                 dataset_name: str = None,
                 sheet_index: int = 0,
                 class_row: int = 0,
                 name_col: int = 0,
                 start_row: int = 1,
                 start_col: int = 1,
                 skip_invalid_rows=True,
                 progress_callback: Callable[[float], None] = None,
                 logger: logging.Logger = None) -> Union[Dataset, None]:
    """
    Try to load the grain size dataset from a file.

    A data files of any types can be regarded as the table that contains rows and columns.

    For the file which contains grain size distributions, it should be the following format:
        * The first valid row should be the headers (i.e. the grain size classes in microns).
        * The following valid rows should be the distributions of samples corresponding to the grain size classes.
        * The first valid column should be the names of samples.
    To make it more flexible, you can use the settings to control the data loader.

    :param filename: The file path which contains the extension.
    :param dataset_name: The name of this dataset. If not indicate it, it will use the filename.
    :param sheet_index: If it is an Excel file (`*.xls` or `*.xlsx`), please indicate the sheet index,
        otherwise it will use the first sheet.
    :param class_row: The row index of the grain size classes.
    :param name_col: The column index of the sample names.
    :param start_row: The start row index of the grain size distributions.
    :param start_col: The start column index of the grain size distributions.
    :param skip_invalid_rows: If `True`, it will skip the invalid rows, rather than break up.
    :param progress_callback: The callable object which will be called at each step to report the progress.
    :param logger: The logger which will be used to log the information of loading.
    :return: If there is any exception raised, it will return `None`, else return a `Dataset` object.
    """
    assert isinstance(filename, str)
    assert len(filename) != 0
    if dataset_name is None:
        dataset_name = os.path.splitext(os.path.basename(filename))[0]
    else:
        assert isinstance(dataset_name, str)
        assert len(dataset_name) != 0
    assert isinstance(sheet_index, int)
    assert sheet_index >= 0
    check_layout(class_row, name_col, start_row, start_col)
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)

    file_type = get_file_type(filename)
    try:
        raw_table = _get_raw_table(
            filename, file_type, sheet_index,
            progress_callback=None if progress_callback is None else lambda p: progress_callback(p * 0.5))
        class_values = raw_table[class_row][start_col:]
        valid, array_or_msg = validate_classes(class_values)
        if valid:
            classes = array_or_msg
            logger.debug(f"Grain size classes in μm: [{','.join([f'{x: 0.4f}' for x in classes])}].")
        else:
            logger.error(f"The assigned series of grain size classes is invalid. {array_or_msg}")
            return None
        sample_names = []
        distributions = []
        for row, row_values in enumerate(raw_table[start_row:], start_row + 1):
            # check if it's an empty row, i.e. the values all are empty string
            is_empty_row = True
            for frequency in row_values[start_col:]:
                if frequency is not None and frequency != "":
                    is_empty_row = False
                    break
            # if it's an empty row, jump this row to process the next one
            if is_empty_row:
                logger.warning(f"Row {row} is empty, skip to next row.")
                continue

            sample_name = row_values[name_col]
            if sample_name is None:
                sample_name = "NONE"
                logger.warning(f"The sample name at row {row} is `None`, use 'NONE' instead.")
            elif not isinstance(sample_name, str):
                sample_name = str(sample_name)
                logger.warning(f"The sample name at row {row} is not `str`, use `str` to covert it.")
            elif len(sample_name) == 0:
                sample_name = "EMPTY"
                logger.warning(f"The sample name at row {row} is empty, use `EMPTY` to covert it.")

            try:
                distribution = np.array(row_values[start_col:], dtype=np.float64)
            except ValueError as e:
                logger.error(f"Can not convert the frequencies at row {row} to a numerical array, "
                             f"it may contains invalid values (e.g. text or empty cell). {e}")
                if skip_invalid_rows:
                    continue
                else:
                    return None
            sample_names.append(sample_name)
            distributions.append(distribution)
            # logger.debug(f"The validation of sample [{sample_name}] at row [{current_row}] is passed.")
            if progress_callback is not None:
                progress = row / len(raw_table) * 0.5 + 0.5
                progress_callback(progress)

        valid, array_or_msg = validate_distributions(distributions)
        if valid:
            distributions = array_or_msg
            dataset = Dataset(dataset_name, sample_names, classes, distributions)
            if progress_callback is not None:
                progress_callback(1.0)
            logger.info("This dataset has been loaded successfully.")
            return dataset
        else:
            logger.error(f"There is at least one grain size distribution is invalid. {array_or_msg}")
    except IOError as e:
        logger.error(f"Can not open this file. {e}")
        return None
    except IndexError as e:
        logger.error(f"The row or column index is out of range, please check. {e}")
        return None



__all__ = ["built_in_losses", "built_in_optimizers", "try_ssu", "try_dataset"]

import logging
import multiprocessing
import time
from typing import *

import numpy as np
from numpy import ndarray
from scipy.optimize import basinhopping, minimize

from .models import DistributionType, Dataset, Sample, SSUResult, ArtificialSample, ArtificialDataset
from .distributions import get_distribution, get_sorted_indexes
from .metrics import loss_numpy

# "cosine" metric has problem
built_in_losses = (
    "1-norm", "2-norm", "3-norm", "4-norm",
    "mae", "mse", "rmse", "rmlse", "lmse", "angular", "cosine")


def check_loss(loss: str):
    assert isinstance(loss, str)
    assert loss in built_in_losses


built_in_optimizers = (
    "Nelder-Mead", "Powell", "CG", "BFGS",
    "L-BFGS-B", "TNC", "SLSQP")


def check_optimizer(optimizer: str):
    assert isinstance(optimizer, str)
    assert optimizer in built_in_optimizers


def try_ssu(sample: Union[ArtificialSample, Sample], distribution_type: DistributionType, n_components: int,
            x0: ndarray = None, loss: str = "lmse", optimizer: str = "L-BFGS-B", try_global: bool = False,
            global_max_niter: int = 100, global_niter_success: int = 5, global_step_size: float = 0.2,
            optimizer_max_niter: int = 10000, need_history: bool = True, logger: logging.Logger = None,
            progress_callback: Callable[[float], None] = None) -> Tuple[Optional[SSUResult], str]:
    assert isinstance(sample, (ArtificialSample, Sample))
    assert isinstance(distribution_type, DistributionType)
    assert isinstance(n_components, int)
    distribution_class = get_distribution(distribution_type)
    if x0 is None:
        x0 = distribution_class.get_defaults(n_components)
    else:
        x0 = np.array(x0)
        assert x0.ndim == 2
        assert x0.shape == (distribution_class.N_PARAMETERS + 1, n_components)
    check_loss(loss)
    loss_func = loss_numpy(loss)
    check_optimizer(optimizer)
    assert isinstance(global_max_niter, int)
    assert isinstance(global_niter_success, int)
    assert isinstance(global_step_size, (int, float))
    assert isinstance(optimizer_max_niter, int)
    assert global_max_niter > 0
    assert global_niter_success > 0
    assert global_step_size > 0.0
    assert optimizer_max_niter > 0
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)

    start_text = f"""Performing the SSU algorithm on the sample ({sample.name}).
    Distribution type: {distribution_type.name}
    Number of components: {n_components}
    x0: {x0 if x0 is None else x0.tolist()}
    Loss: {loss}
    Optimizer: {optimizer}
    Try global optimization: {try_global}
    Global maximum number of iterations: {global_max_niter}
    Global number of success iterations: {global_niter_success}
    Global step size: {global_step_size}
    Optimizer maximum number of iterations: {optimizer_max_niter}
    Need history: {need_history}"""
    logger.debug(start_text)

    start_time = time.time()
    global_iteration = 0
    iteration = 0
    max_iterations = global_max_niter * optimizer_max_niter if try_global else optimizer_max_niter
    history = [np.expand_dims(x0, axis=0)]
    classes = np.expand_dims(np.expand_dims(sample.classes_phi, 0), 0).repeat(n_components, 1)

    def closure(x):
        x = x.reshape((1, distribution_class.N_PARAMETERS + 1, n_components))
        proportions, components, _ = distribution_class.interpret(x, classes, sample.interval_phi)
        pred_distribution = (proportions[0] @ components[0]).squeeze()
        return loss_func(pred_distribution, sample.distribution, None)

    def callback(x: ndarray):
        nonlocal iteration
        iteration += 1
        x = x.reshape((1, distribution_class.N_PARAMETERS + 1, n_components))
        progress = iteration / max_iterations
        if need_history:
            history.append(x)
        if progress_callback is not None:
            progress_callback(progress)

    def global_callback(x: ndarray, f: float, accept: bool):
        nonlocal global_iteration
        global_iteration += 1
        x = x.reshape((1, distribution_class.N_PARAMETERS + 1, n_components))
        logger.debug(f"The global epoch {global_iteration} finished, x: {x}, function value: {f}, accepted: {accept}.")

    if try_global:
        global_result = basinhopping(
            closure, x0=x0, minimizer_kwargs=dict(
                method=optimizer, callback=callback, options=dict(maxiter=optimizer_max_niter)),
            niter_success=global_niter_success, niter=global_max_niter,
            stepsize=global_step_size, callback=global_callback)
        if global_result.lowest_optimization_result.success or global_result.lowest_optimization_result.status == 9:
            parameters = np.reshape(global_result.x, (1, distribution_class.N_PARAMETERS + 1, n_components))
            message = global_result.message
        else:
            logger.error(f"The fitting process terminated with a error: {global_result.message}.")
            return None, global_result.message
    else:
        local_result = minimize(closure, x0=x0, method=optimizer,
                                callback=callback, options=dict(maxiter=optimizer_max_niter))
        if local_result.success or local_result.status == 9:
            parameters = np.reshape(local_result.x, (1, distribution_class.N_PARAMETERS + 1, n_components))
            message = local_result.message
        else:
            logger.error(f"The fitting process terminated with a error: {local_result.message}.")
            return None, local_result.message

    time_spent = time.time() - start_time
    # sort the components by their grain sizes (from fine to coarse)
    sorted_indexes = get_sorted_indexes(distribution_type, parameters, classes, sample.interval_phi)
    if need_history:
        parameters = np.concatenate(history, axis=0)
    sorted_parameters = np.zeros_like(parameters)
    for i, j in enumerate(sorted_indexes):
        sorted_parameters[:, :, i] = parameters[:, :, j]
    settings = dict(loss=loss, optimizer=optimizer, try_global=try_global, global_max_niter=global_max_niter,
                    global_niter_success=global_max_niter, global_step_size=global_step_size,
                    optimizer_max_niter=optimizer_max_niter, need_history=need_history)
    ssu_result = SSUResult(sample, distribution_type, sorted_parameters, time_spent, x0=x0, settings=settings)
    logger.debug(f"The fitting process successfully finished. {message}")
    return ssu_result, message


def _execute(args: Tuple[Union[ArtificialSample, Sample], DistributionType, int, Dict[str, Any]]):
    sample, distribution_type, n_components, options = args
    return try_ssu(sample, distribution_type, n_components, **options)


def try_dataset(
        dataset: Union[ArtificialDataset, Dataset],
        distribution_type: DistributionType,
        n_components: int,
        n_processes: int = 1,
        options: Dict[str, Any] = None):
    multiprocessing.freeze_support()
    pool = multiprocessing.Pool(n_processes)
    args = [(sample, distribution_type, n_components, options) for sample in dataset]
    results = pool.map(_execute, args)
    succeeded_results: List[SSUResult] = []
    failed_samples: List[Tuple[int, str]] = []
    for i, (result, message) in enumerate(results):
        if isinstance(result, SSUResult):
            succeeded_results.append(result)
        else:
            failed_samples.append((i, message))
    return succeeded_results, failed_samples



__all__ = ["UDMResult"]

import copy
from typing import *

import numpy as np
from numpy import ndarray

from ..models import DistributionType, KernelType, ArtificialDataset, Dataset
from ..distributions import get_distribution, get_sorted_indexes
from ..metrics import loss_numpy


class UDMResult:
    def __init__(self, dataset: [ArtificialDataset, Dataset], kernel_type: KernelType, n_components: int,
                 parameters: ndarray, time_spent: Union[int, float], x0: ndarray = None,
                 settings: Dict[str, Any] = None, loss_series: Dict[str, ndarray] = None):
        self._dataset = dataset
        self._kernel_type = kernel_type
        self._n_components = n_components
        self._time_spent = time_spent
        self._x0 = x0
        self._settings = settings
        self._loss_series = loss_series
        self._classes_phi = np.expand_dims(np.expand_dims(self.dataset.classes_phi, axis=0), axis=0).repeat(
            self.n_samples, axis=0).repeat(self.n_components, axis=1)
        self._interval_phi = np.abs((self.dataset.classes_phi[0] - self.dataset.classes_phi[-1]) / (self.n_classes - 1))
        indexes = get_sorted_indexes(self.distribution_type, parameters[-1], self._classes_phi, self._interval_phi)
        sorted_parameters = np.zeros_like(parameters)
        for i, j in enumerate(indexes):
            sorted_parameters[:, :, :, i] = parameters[:, :, :, j]
        self._parameters = sorted_parameters
        self._update(-1)

    @property
    def dataset(self) -> Union[ArtificialDataset, Dataset]:
        return self._dataset

    @property
    def n_samples(self) -> int:
        return len(self.dataset)

    @property
    def n_components(self) -> int:
        return self._n_components

    @property
    def n_classes(self) -> int:
        return len(self.dataset.classes)

    @property
    def kernel_type(self) -> KernelType:
        return self._kernel_type

    @property
    def distribution_type(self) -> DistributionType:
        return DistributionType.__members__[self._kernel_type.name]

    @property
    def proportions(self) -> ndarray:
        return self._proportions

    @property
    def components(self) -> ndarray:
        return self._components

    @property
    def distributions(self) -> ndarray:
        return (self._proportions @ self._components).squeeze(1)

    @property
    def time_spent(self) -> Union[int, float]:
        return self._time_spent

    @property
    def x0(self) -> Optional[ndarray]:
        return self._x0

    @property
    def n_iterations(self) -> int:
        return self._parameters.shape[0]

    @property
    def parameters(self) -> ndarray:
        return self._parameters

    @property
    def history(self):
        for i in range(self._parameters.shape[0]):
            copy_result = copy.copy(self)
            copy_result._update(i)
            yield copy_result

    @property
    def settings(self) -> Optional[Dict[str, Any]]:
        if self._settings is None:
            return None
        else:
            return self._settings.copy()

    def _update(self, index: int):
        distribution_class = get_distribution(DistributionType.__members__[self.kernel_type.name])
        proportions, components, _ = distribution_class.interpret(
            self._parameters[index], self._classes_phi, self._interval_phi)
        proportions[np.logical_or(np.isnan(proportions), np.isinf(proportions))] = 0.0
        components[np.logical_or(np.isnan(components), np.isinf(components))] = 0.0
        self._proportions = proportions
        self._components = components

    def loss(self, name: str) -> float:
        loss_func = loss_numpy(name)
        observation = self._dataset.distributions
        prediction = (self._proportions @ self._components).squeeze(1)
        return loss_func(prediction, observation, None)

    def loss_series(self, name: str) -> ndarray:
        if name in self._loss_series:
            return self._loss_series[name]
        series = []
        loss_func = loss_numpy(name)
        observation = self._dataset.distributions
        for result in self.history:
            prediction = (result._proportions @ result._components).squeeze(1)
            loss = loss_func(prediction, observation, None)
            series.append(loss)
        series = np.array(series)
        self._loss_series[name] = series
        return series

    def class_wise_losses(self, name: str) -> ndarray:
        loss_func = loss_numpy(name)
        observation = self._dataset.distributions
        prediction = (self._proportions @ self._components).squeeze(1)
        losses = loss_func(prediction, observation, 0)
        return losses

    def sample_wise_losses(self, name: str) -> ndarray:
        loss_func = loss_numpy(name)
        observation = self._dataset.distributions
        prediction = (self._proportions @ self._components).squeeze(1)
        losses = loss_func(prediction, observation, 1)
        return losses



__all__ = ["BoxplotChart"]

from typing import *

import matplotlib.pyplot as plt
from numpy import ndarray

from . import normal_color, BaseChart


class BoxplotChart(BaseChart):
    def __init__(self, parent=None, figsize=(3, 2.5)):
        super().__init__(parent=parent, figsize=figsize)
        self._axes = self._figure.subplots()
        self.setWindowTitle(self.tr("Boxplot Chart"))
        self._last_result = None

    def show_dataset(self, dataset: Sequence[ndarray], xlabels: Sequence[str], ylabel: str, title: str = ""):
        self._axes.clear()
        assert len(dataset) == len(xlabels)
        # "whiskers", "caps", "boxes", "medians", "fliers", "means"
        artists = self._axes.boxplot(dataset, labels=xlabels, patch_artist=True)
        cmap = plt.get_cmap()
        for i, box in enumerate(artists["boxes"]):
            box.set_facecolor(cmap(i))
            box.set_edgecolor(normal_color())
        for whisker in artists["whiskers"]:
            whisker.set_color(normal_color())
        for median in artists["medians"]:
            median.set_color(normal_color())
        for cap in artists["caps"]:
            cap.set_color(normal_color())
        for i, flier in enumerate(artists["fliers"]):
            flier.set_markerfacecolor(cmap(i))
            flier.set_markeredgewidth(0.0)
        self._axes.set_ylabel(ylabel)
        self._axes.set_title(title)
        self._figure.tight_layout()
        self._canvas.draw()
        self._last_result = dataset, xlabels, ylabel, title

    def update_chart(self):
        if self._last_result is not None:
            self._figure.clear()
            self._axes = self._figure.subplots()
            self.show_dataset(*self._last_result)

    def retranslate(self):
        self.setWindowTitle(self.tr("Boxplot Chart"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))



__all__ = ["built_in_losses", "try_emma"]

import logging
import time
from typing import *

import numpy as np
import torch
from numpy import ndarray

from .models import KernelType, ArtificialDataset, Dataset, EMMAResult
from .kernels import loss_torch, ProportionModule, get_kernel

built_in_losses = (
    "1-norm", "2-norm", "3-norm", "4-norm",
    "mae", "mse", "rmse", "rmlse", "lmse", "angular", "cosine")

torch.set_default_dtype(torch.float64)


class EMMAModule(torch.nn.Module):
    def __init__(self, n_samples: int, n_members: int, classes_phi: np.ndarray,
                 kernel_type: KernelType, x0: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = len(classes_phi)
        self._interval_phi = np.abs((classes_phi[0] - classes_phi[-1]) / (classes_phi.shape[0] - 1))
        self._classes_phi = torch.nn.Parameter(
            torch.from_numpy(classes_phi).repeat(1, n_members, 1), requires_grad=False)
        self.kernel_type = kernel_type
        self.proportions = ProportionModule(n_samples, n_members)
        self.end_members = get_kernel(kernel_type, 1, self.n_members, self.n_classes, x0)

    def forward(self) -> Tuple[torch.Tensor, torch.Tensor]:
        # n_samples x n_members
        proportions = self.proportions().squeeze(1)
        # n_members x n_classes
        end_members = self.end_members(self._classes_phi, self._interval_phi).squeeze(0)
        # n_samples x n_classes
        # distributions = proportions @ end_members
        return proportions, end_members


def try_emma(dataset: Union[ArtificialDataset, Dataset], kernel_type: KernelType, n_members: int, x0: ndarray = None,
             device="cpu", loss="lmse", pretrain_epochs=0, min_epochs=100, max_epochs=10000,
             precision: Union[int, float] = 6, learning_rate=5e-3, betas=(0.8, 0.5), update_end_members=True,
             need_history=True, logger: logging.Logger = None, progress_callback: Callable[[float], None] = None):
    assert isinstance(dataset, (ArtificialDataset, Dataset))
    assert isinstance(kernel_type, KernelType)
    assert isinstance(n_members, int)
    if x0 is not None:
        assert isinstance(x0, ndarray)
        assert x0.ndim == 2
        assert x0.shape[1] == n_members
        x0 = x0.astype(np.float64)
    available_devices = ["cpu"]
    if torch.cuda.is_available():
        available_devices.append("cuda")
    available_devices.extend([f"cuda:{i}" for i in range(torch.cuda.device_count())])
    if "cuda:0" in available_devices:
        available_devices.remove("cuda:0")
    assert device in available_devices
    assert loss in built_in_losses
    assert isinstance(pretrain_epochs, int)
    assert isinstance(min_epochs, int)
    assert isinstance(max_epochs, int)
    assert isinstance(precision, (int, float))
    assert isinstance(learning_rate, float)
    assert isinstance(betas, tuple)
    assert len(betas) == 2
    beta1, beta2 = betas
    assert isinstance(beta1, float)
    assert isinstance(beta2, float)
    assert pretrain_epochs >= 0
    assert min_epochs > 0
    assert max_epochs > 0
    assert 1.0 < precision < 100.0
    assert learning_rate > 0.0
    assert 0.0 < beta1 < 1.0
    assert 0.0 < beta2 < 1.0
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)

    start_text = f"""Performing the EMMA algorithm on the dataset ({dataset.name}).
    Kernel type: {kernel_type.name}
    Number of end members: {n_members}
    x0: {x0 if x0 is None else x0.tolist()}
    Device: {device}
    Loss: {loss}
    Pretrain epochs: {pretrain_epochs}
    Minimum epochs: {min_epochs}
    Maximum epochs: {max_epochs}
    Precision decimals (10^-x): {precision}
    Learning rate: {learning_rate}
    Betas of weight decay: {betas}
    Update end members: {update_end_members}
    Need history: {need_history}"""
    logger.debug(start_text)

    observation = torch.from_numpy(dataset.distributions.astype(np.float64)).to(device)
    emma = EMMAModule(len(dataset), n_members, dataset.classes_phi.astype(np.float64), kernel_type, x0).to(device)
    loss_func = loss_torch(loss)
    optimizer = torch.optim.Adam(emma.parameters(), lr=learning_rate, betas=betas)
    total_loss_series = []
    history: List[Tuple[ndarray, ndarray]] = []
    emma.end_members.requires_grad_(False)
    max_total_epochs = pretrain_epochs + max_epochs
    start = time.time()
    for pretrain_epoch in range(pretrain_epochs):
        proportions, end_members = emma()
        prediction = proportions @ end_members
        loss_i = loss_func(prediction, observation)
        total_loss_series.append(loss_i.item())
        optimizer.zero_grad()
        loss_i.backward()
        torch.nn.utils.clip_grad_norm_(emma.parameters(), 1e-1)
        optimizer.step()
        if need_history:
            history.append((proportions.detach().cpu().numpy(), end_members.detach().cpu().numpy()))
        if progress_callback is not None:
            progress_callback(pretrain_epoch / max_total_epochs)

    emma.end_members.requires_grad_(update_end_members)
    for epoch in range(max_epochs):
        # train
        proportions, end_members = emma()
        prediction = proportions @ end_members
        loss_i = loss_func(prediction, observation)
        if np.isnan(loss_i.item()):
            logger.warning("Loss is NaN, training has beem terminated.")
            break
        total_loss_series.append(loss_i.item())
        optimizer.zero_grad()
        loss_i.backward()
        torch.nn.utils.clip_grad_norm_(emma.parameters(), 1e-1)
        optimizer.step()
        if need_history:
            history.append((proportions.detach().cpu().numpy(), end_members.detach().cpu().numpy()))
        if progress_callback is not None:
            progress_callback((pretrain_epochs+epoch) / max_total_epochs)
        if epoch > min_epochs:
            delta_loss = np.mean(total_loss_series[-100:-80])-np.mean(total_loss_series[-20:])
            if delta_loss < 10**(-precision):
                break

    # algorithm finished, preparing the result
    if device[:4] == "cuda":
        torch.cuda.synchronize()
    time_spent = time.time() - start
    settings = dict(device=device, loss=loss, pretrain_epochs=pretrain_epochs, min_epochs=min_epochs,
                    max_epochs=max_epochs, precision=precision, learning_rate=learning_rate,
                    betas=betas, update_end_members=update_end_members, need_history=need_history)
    total_loss_series = np.array(total_loss_series)
    loss_series = {loss: total_loss_series}
    if need_history:
        proportions = np.concatenate([np.expand_dims(proportions, axis=0) for proportions, _ in history], axis=0)
        end_members = np.concatenate([np.expand_dims(end_members, axis=0) for _, end_members in history], axis=0)
    else:
        with torch.no_grad():
            proportions, end_members = emma()
            proportions = np.expand_dims(proportions.cpu().numpy(), axis=0)
            end_members = np.expand_dims(end_members.cpu().numpy(), axis=0)
    result = EMMAResult(dataset, kernel_type, n_members, proportions, end_members, time_spent, x0, settings, loss_series)
    if progress_callback is not None:
        progress_callback(1.0)
    return result



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qgrain.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cqgrain.proto\"=\n\x06Sample\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x63lasses\x18\x02 \x03(\x02\x12\x14\n\x0c\x64istribution\x18\x03 \x03(\x02\"{\n\x07\x44\x61taset\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tn_samples\x18\x02 \x01(\x05\x12\x11\n\tn_classes\x18\x03 \x01(\x05\x12\x14\n\x0csample_names\x18\x04 \x03(\t\x12\x0f\n\x07\x63lasses\x18\x05 \x03(\x02\x12\x15\n\rdistributions\x18\x06 \x01(\x0c\"\x15\n\x13ServiceStateRequest\"|\n\x14ServiceStateResponse\x12\x13\n\x0bmax_workers\x18\x01 \x01(\x05\x12\x1a\n\x12max_message_length\x18\x02 \x01(\x05\x12\x19\n\x11\x61vailable_devices\x18\t \x03(\t\x12\x18\n\x10max_dataset_size\x18\n \x01(\x05\"/\n\x12StatisticalRequest\x12\x19\n\x07\x64\x61taset\x18\x01 \x01(\x0b\x32\x08.Dataset\"\x8f\x03\n\x15StatisticalParameters\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x0c\n\x04mean\x18\x02 \x01(\x02\x12\x0b\n\x03std\x18\x03 \x01(\x02\x12\x10\n\x08skewness\x18\x04 \x01(\x02\x12\x10\n\x08kurtosis\x18\x05 \x01(\x02\x12\x13\n\x06median\x18\x06 \x01(\x02H\x00\x88\x01\x01\x12\x11\n\x04mode\x18\x07 \x01(\x02H\x01\x88\x01\x01\x12\r\n\x05modes\x18\x08 \x03(\x02\x12\x1d\n\x10mean_description\x18\t \x01(\tH\x02\x88\x01\x01\x12\x1c\n\x0fstd_description\x18\n \x01(\tH\x03\x88\x01\x01\x12!\n\x14skewness_description\x18\x0b \x01(\tH\x04\x88\x01\x01\x12!\n\x14kurtosis_description\x18\x0c \x01(\tH\x05\x88\x01\x01\x42\t\n\x07_medianB\x07\n\x05_modeB\x13\n\x11_mean_descriptionB\x12\n\x10_std_descriptionB\x17\n\x15_skewness_descriptionB\x17\n\x15_kurtosis_description\"\xac\x05\n\x11StatisticalResult\x12/\n\narithmetic\x18\x01 \x01(\x0b\x32\x16.StatisticalParametersH\x00\x88\x01\x01\x12.\n\tgeometric\x18\x02 \x01(\x0b\x32\x16.StatisticalParametersH\x01\x88\x01\x01\x12\x30\n\x0blogarithmic\x18\x03 \x01(\x0b\x32\x16.StatisticalParametersH\x02\x88\x01\x01\x12\x33\n\x0egeometric_fw57\x18\x04 \x01(\x0b\x32\x16.StatisticalParametersH\x03\x88\x01\x01\x12\x35\n\x10logarithmic_fw57\x18\x05 \x01(\x0b\x32\x16.StatisticalParametersH\x04\x88\x01\x01\x12\x17\n\x0fproportions_gsm\x18\x06 \x03(\x02\x12\x17\n\x0fproportions_ssc\x18\x07 \x03(\x02\x12\x19\n\x11proportions_bgssc\x18\x08 \x03(\x02\x12\x38\n\x0bproportions\x18\t \x03(\x0b\x32#.StatisticalResult.ProportionsEntry\x12\x19\n\x0cgroup_folk54\x18\n \x01(\tH\x05\x88\x01\x01\x12\x1e\n\x11group_bp12_symbol\x18\x0b \x01(\tH\x06\x88\x01\x01\x12\x17\n\ngroup_bp12\x18\x0c \x01(\tH\x07\x88\x01\x01\x1a\x32\n\x10ProportionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x42\r\n\x0b_arithmeticB\x0c\n\n_geometricB\x0e\n\x0c_logarithmicB\x11\n\x0f_geometric_fw57B\x13\n\x11_logarithmic_fw57B\x0f\n\r_group_folk54B\x14\n\x12_group_bp12_symbolB\r\n\x0b_group_bp12\"K\n\x13StatisticalResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12#\n\x07results\x18\x02 \x03(\x0b\x32\x12.StatisticalResult\"\xbb\x02\n\nSSURequest\x12\x17\n\x06sample\x18\x01 \x01(\x0b\x32\x07.Sample\x12,\n\x11\x64istribution_type\x18\x02 \x01(\x0e\x32\x11.DistributionType\x12\x14\n\x0cn_components\x18\x03 \x01(\x05\x12\x0f\n\x02x0\x18\x04 \x01(\x0cH\x00\x88\x01\x01\x12\x0c\n\x04loss\x18\x05 \x01(\t\x12\x11\n\toptimizer\x18\x06 \x01(\t\x12\x12\n\ntry_global\x18\x07 \x01(\x08\x12\x18\n\x10global_max_niter\x18\x08 \x01(\x05\x12\x1c\n\x14global_niter_success\x18\t \x01(\x05\x12\x18\n\x10global_step_size\x18\n \x01(\x02\x12\x1b\n\x13optimizer_max_niter\x18\x0b \x01(\x05\x12\x14\n\x0cneed_history\x18\x0c \x01(\x08\x42\x05\n\x03_x0\"\xf2\x01\n\x0bSSUResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x17\n\ntime_spent\x18\x02 \x01(\x02H\x00\x88\x01\x01\x12\x19\n\x0cn_iterations\x18\x03 \x01(\x05H\x01\x88\x01\x01\x12\x19\n\x0cn_parameters\x18\x04 \x01(\x05H\x02\x88\x01\x01\x12\x19\n\x0cn_components\x18\x05 \x01(\x05H\x03\x88\x01\x01\x12\x17\n\nparameters\x18\x06 \x01(\x0cH\x04\x88\x01\x01\x42\r\n\x0b_time_spentB\x0f\n\r_n_iterationsB\x0f\n\r_n_parametersB\x0f\n\r_n_componentsB\r\n\x0b_parameters\"\xcb\x02\n\x0b\x45MMARequest\x12\x19\n\x07\x64\x61taset\x18\x01 \x01(\x0b\x32\x08.Dataset\x12,\n\x11\x64istribution_type\x18\x02 \x01(\x0e\x32\x11.DistributionType\x12\x11\n\tn_members\x18\x03 \x01(\x05\x12\x0f\n\x02x0\x18\x04 \x01(\x0cH\x00\x88\x01\x01\x12\x0e\n\x06\x64\x65vice\x18\x05 \x01(\t\x12\x0c\n\x04loss\x18\x06 \x01(\t\x12\x17\n\x0fpretrain_epochs\x18\x07 \x01(\x05\x12\x12\n\nmin_epochs\x18\x08 \x01(\x05\x12\x12\n\nmax_epochs\x18\t \x01(\x05\x12\x11\n\tprecision\x18\n \x01(\x02\x12\x15\n\rlearning_rate\x18\x0b \x01(\x02\x12\r\n\x05\x62\x65tas\x18\x0c \x03(\x02\x12\x1a\n\x12update_end_members\x18\r \x01(\x08\x12\x14\n\x0cneed_history\x18\x0e \x01(\x08\x42\x05\n\x03_x0\"\xc9\x02\n\x0c\x45MMAResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x17\n\ntime_spent\x18\x02 \x01(\x02H\x00\x88\x01\x01\x12\x19\n\x0cn_iterations\x18\x03 \x01(\x05H\x01\x88\x01\x01\x12\x16\n\tn_samples\x18\x04 \x01(\x05H\x02\x88\x01\x01\x12\x16\n\tn_members\x18\x05 \x01(\x05H\x03\x88\x01\x01\x12\x16\n\tn_classes\x18\x06 \x01(\x05H\x04\x88\x01\x01\x12\x18\n\x0bproportions\x18\x07 \x01(\x0cH\x05\x88\x01\x01\x12\x18\n\x0b\x65nd_members\x18\x08 \x01(\x0cH\x06\x88\x01\x01\x12\x0e\n\x06losses\x18\t \x03(\x02\x42\r\n\x0b_time_spentB\x0f\n\r_n_iterationsB\x0c\n\n_n_samplesB\x0c\n\n_n_membersB\x0c\n\n_n_classesB\x0e\n\x0c_proportionsB\x0e\n\x0c_end_members\"\xbd\x02\n\nUDMRequest\x12\x19\n\x07\x64\x61taset\x18\x01 \x01(\x0b\x32\x08.Dataset\x12,\n\x11\x64istribution_type\x18\x02 \x01(\x0e\x32\x11.DistributionType\x12\x14\n\x0cn_components\x18\x03 \x01(\x05\x12\x0f\n\x02x0\x18\x04 \x01(\x0cH\x00\x88\x01\x01\x12\x0e\n\x06\x64\x65vice\x18\x05 \x01(\t\x12\x17\n\x0fpretrain_epochs\x18\x06 \x01(\x05\x12\x12\n\nmin_epochs\x18\x07 \x01(\x05\x12\x12\n\nmax_epochs\x18\x08 \x01(\x05\x12\x11\n\tprecision\x18\t \x01(\x02\x12\x15\n\rlearning_rate\x18\n \x01(\x02\x12\r\n\x05\x62\x65tas\x18\x0b \x03(\x02\x12\x18\n\x10\x63onstraint_level\x18\x0c \x01(\x02\x12\x14\n\x0cneed_history\x18\r \x01(\x08\x42\x05\n\x03_x0\"\xdf\x02\n\x0bUDMResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x17\n\ntime_spent\x18\x02 \x01(\x02H\x00\x88\x01\x01\x12\x19\n\x0cn_iterations\x18\x03 \x01(\x05H\x01\x88\x01\x01\x12\x16\n\tn_samples\x18\x04 \x01(\x05H\x02\x88\x01\x01\x12\x19\n\x0cn_components\x18\x05 \x01(\x05H\x03\x88\x01\x01\x12\x16\n\tn_classes\x18\x06 \x01(\x05H\x04\x88\x01\x01\x12\x17\n\nparameters\x18\x07 \x01(\x0cH\x05\x88\x01\x01\x12\x1b\n\x13\x64istribution_losses\x18\x08 \x03(\x02\x12\x18\n\x10\x63omponent_losses\x18\t \x03(\x02\x12\x14\n\x0ctotal_losses\x18\n \x03(\x02\x42\r\n\x0b_time_spentB\x0f\n\r_n_iterationsB\x0c\n\n_n_samplesB\x0f\n\r_n_componentsB\x0c\n\n_n_classesB\r\n\x0b_parameters*b\n\x10\x44istributionType\x12\x11\n\rNonparametric\x10\x00\x12\n\n\x06Normal\x10\x01\x12\x0e\n\nSkewNormal\x10\x02\x12\x0b\n\x07Weibull\x10\x03\x12\x12\n\x0eGeneralWeibull\x10\x04\x32\x9b\x02\n\x06QGrain\x12\x42\n\x11get_service_state\x12\x14.ServiceStateRequest\x1a\x15.ServiceStateResponse\"\x00\x12=\n\x0eget_statistics\x12\x13.StatisticalRequest\x1a\x14.StatisticalResponse\"\x00\x12-\n\x0eget_ssu_result\x12\x0b.SSURequest\x1a\x0c.SSUResponse\"\x00\x12\x30\n\x0fget_emma_result\x12\x0c.EMMARequest\x1a\r.EMMAResponse\"\x00\x12-\n\x0eget_udm_result\x12\x0b.UDMRequest\x1a\x0c.UDMResponse\"\x00\x62\x06proto3')

_DISTRIBUTIONTYPE = DESCRIPTOR.enum_types_by_name['DistributionType']
DistributionType = enum_type_wrapper.EnumTypeWrapper(_DISTRIBUTIONTYPE)
Nonparametric = 0
Normal = 1
SkewNormal = 2
Weibull = 3
GeneralWeibull = 4


_SAMPLE = DESCRIPTOR.message_types_by_name['Sample']
_DATASET = DESCRIPTOR.message_types_by_name['Dataset']
_SERVICESTATEREQUEST = DESCRIPTOR.message_types_by_name['ServiceStateRequest']
_SERVICESTATERESPONSE = DESCRIPTOR.message_types_by_name['ServiceStateResponse']
_STATISTICALREQUEST = DESCRIPTOR.message_types_by_name['StatisticalRequest']
_STATISTICALPARAMETERS = DESCRIPTOR.message_types_by_name['StatisticalParameters']
_STATISTICALRESULT = DESCRIPTOR.message_types_by_name['StatisticalResult']
_STATISTICALRESULT_PROPORTIONSENTRY = _STATISTICALRESULT.nested_types_by_name['ProportionsEntry']
_STATISTICALRESPONSE = DESCRIPTOR.message_types_by_name['StatisticalResponse']
_SSUREQUEST = DESCRIPTOR.message_types_by_name['SSURequest']
_SSURESPONSE = DESCRIPTOR.message_types_by_name['SSUResponse']
_EMMAREQUEST = DESCRIPTOR.message_types_by_name['EMMARequest']
_EMMARESPONSE = DESCRIPTOR.message_types_by_name['EMMAResponse']
_UDMREQUEST = DESCRIPTOR.message_types_by_name['UDMRequest']
_UDMRESPONSE = DESCRIPTOR.message_types_by_name['UDMResponse']
Sample = _reflection.GeneratedProtocolMessageType('Sample', (_message.Message,), {
  'DESCRIPTOR' : _SAMPLE,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:Sample)
  })
_sym_db.RegisterMessage(Sample)

Dataset = _reflection.GeneratedProtocolMessageType('Dataset', (_message.Message,), {
  'DESCRIPTOR' : _DATASET,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:Dataset)
  })
_sym_db.RegisterMessage(Dataset)

ServiceStateRequest = _reflection.GeneratedProtocolMessageType('ServiceStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVICESTATEREQUEST,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:ServiceStateRequest)
  })
_sym_db.RegisterMessage(ServiceStateRequest)

ServiceStateResponse = _reflection.GeneratedProtocolMessageType('ServiceStateResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVICESTATERESPONSE,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:ServiceStateResponse)
  })
_sym_db.RegisterMessage(ServiceStateResponse)

StatisticalRequest = _reflection.GeneratedProtocolMessageType('StatisticalRequest', (_message.Message,), {
  'DESCRIPTOR' : _STATISTICALREQUEST,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:StatisticalRequest)
  })
_sym_db.RegisterMessage(StatisticalRequest)

StatisticalParameters = _reflection.GeneratedProtocolMessageType('StatisticalParameters', (_message.Message,), {
  'DESCRIPTOR' : _STATISTICALPARAMETERS,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:StatisticalParameters)
  })
_sym_db.RegisterMessage(StatisticalParameters)

StatisticalResult = _reflection.GeneratedProtocolMessageType('StatisticalResult', (_message.Message,), {

  'ProportionsEntry' : _reflection.GeneratedProtocolMessageType('ProportionsEntry', (_message.Message,), {
    'DESCRIPTOR' : _STATISTICALRESULT_PROPORTIONSENTRY,
    '__module__' : 'qgrain_pb2'
    # @@protoc_insertion_point(class_scope:StatisticalResult.ProportionsEntry)
    })
  ,
  'DESCRIPTOR' : _STATISTICALRESULT,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:StatisticalResult)
  })
_sym_db.RegisterMessage(StatisticalResult)
_sym_db.RegisterMessage(StatisticalResult.ProportionsEntry)

StatisticalResponse = _reflection.GeneratedProtocolMessageType('StatisticalResponse', (_message.Message,), {
  'DESCRIPTOR' : _STATISTICALRESPONSE,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:StatisticalResponse)
  })
_sym_db.RegisterMessage(StatisticalResponse)

SSURequest = _reflection.GeneratedProtocolMessageType('SSURequest', (_message.Message,), {
  'DESCRIPTOR' : _SSUREQUEST,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:SSURequest)
  })
_sym_db.RegisterMessage(SSURequest)

SSUResponse = _reflection.GeneratedProtocolMessageType('SSUResponse', (_message.Message,), {
  'DESCRIPTOR' : _SSURESPONSE,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:SSUResponse)
  })
_sym_db.RegisterMessage(SSUResponse)

EMMARequest = _reflection.GeneratedProtocolMessageType('EMMARequest', (_message.Message,), {
  'DESCRIPTOR' : _EMMAREQUEST,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:EMMARequest)
  })
_sym_db.RegisterMessage(EMMARequest)

EMMAResponse = _reflection.GeneratedProtocolMessageType('EMMAResponse', (_message.Message,), {
  'DESCRIPTOR' : _EMMARESPONSE,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:EMMAResponse)
  })
_sym_db.RegisterMessage(EMMAResponse)

UDMRequest = _reflection.GeneratedProtocolMessageType('UDMRequest', (_message.Message,), {
  'DESCRIPTOR' : _UDMREQUEST,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:UDMRequest)
  })
_sym_db.RegisterMessage(UDMRequest)

UDMResponse = _reflection.GeneratedProtocolMessageType('UDMResponse', (_message.Message,), {
  'DESCRIPTOR' : _UDMRESPONSE,
  '__module__' : 'qgrain_pb2'
  # @@protoc_insertion_point(class_scope:UDMResponse)
  })
_sym_db.RegisterMessage(UDMResponse)

_QGRAIN = DESCRIPTOR.services_by_name['QGrain']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _STATISTICALRESULT_PROPORTIONSENTRY._options = None
  _STATISTICALRESULT_PROPORTIONSENTRY._serialized_options = b'8\001'
  _DISTRIBUTIONTYPE._serialized_start=3471
  _DISTRIBUTIONTYPE._serialized_end=3569
  _SAMPLE._serialized_start=16
  _SAMPLE._serialized_end=77
  _DATASET._serialized_start=79
  _DATASET._serialized_end=202
  _SERVICESTATEREQUEST._serialized_start=204
  _SERVICESTATEREQUEST._serialized_end=225
  _SERVICESTATERESPONSE._serialized_start=227
  _SERVICESTATERESPONSE._serialized_end=351
  _STATISTICALREQUEST._serialized_start=353
  _STATISTICALREQUEST._serialized_end=400
  _STATISTICALPARAMETERS._serialized_start=403
  _STATISTICALPARAMETERS._serialized_end=802
  _STATISTICALRESULT._serialized_start=805
  _STATISTICALRESULT._serialized_end=1489
  _STATISTICALRESULT_PROPORTIONSENTRY._serialized_start=1300
  _STATISTICALRESULT_PROPORTIONSENTRY._serialized_end=1350
  _STATISTICALRESPONSE._serialized_start=1491
  _STATISTICALRESPONSE._serialized_end=1566
  _SSUREQUEST._serialized_start=1569
  _SSUREQUEST._serialized_end=1884
  _SSURESPONSE._serialized_start=1887
  _SSURESPONSE._serialized_end=2129
  _EMMAREQUEST._serialized_start=2132
  _EMMAREQUEST._serialized_end=2463
  _EMMARESPONSE._serialized_start=2466
  _EMMARESPONSE._serialized_end=2795
  _UDMREQUEST._serialized_start=2798
  _UDMREQUEST._serialized_end=3115
  _UDMRESPONSE._serialized_start=3118
  _UDMRESPONSE._serialized_end=3469
  _QGRAIN._serialized_start=3572
  _QGRAIN._serialized_end=3855
# @@protoc_insertion_point(module_scope)



__all__ = ["mode", "modes", "to_cumulative", "reversed_phi_ppf", "interval_phi", "to_phi", "to_microns", "cm", "ss",
           "arithmetic", "geometric", "logarithmic", "logarithmic_fw57", "geometric_fw57", "scale_description",
           "proportions_gsm", "proportions_ssc", "proportions_bgssc", "all_proportions",
           "group_gsm_folk54", "group_ssc_folk54", "group_folk54", "GROUP_BP12_SYMBOL_MAP",
           "group_gsm_bp12", "group_ssc_bp12", "group_bp12", "major_statistics", "all_statistics"]

import string
from typing import *

import numpy as np
from numpy import ndarray
from scipy.interpolate import interp1d
from scipy.signal import find_peaks


def mode(classes: ndarray,
         classes_phi: ndarray,
         distribution: ndarray,
         is_geometric: bool = False) -> float:
    """
    Get the mode size of a grain size distribution.

    :param classes: The grain size classes in microns.
    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :param is_geometric: If it‘s `True`, the returned mode size is in microns, else it's in phi values.
    :return: The mode size of a grain size distribution.
    """
    max_pos = np.unravel_index(np.argmax(distribution), distribution.shape)
    if is_geometric:
        return classes[max_pos]
    else:
        return classes_phi[max_pos]


def modes(classes: ndarray,
          classes_phi: ndarray,
          distribution: ndarray,
          is_geometric=False, trace=0.01) -> \
        Tuple[Sequence[float], Sequence[float]]:
    """
    Get the modes of a grain size distribution. **The number of modes may be greater than 1.**

    :param classes: The grain size classes in microns.
    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :param is_geometric: If it's `True`, the returned mode size is in microns, else it's in phi values.
    :param trace: If the peak frequency is less than `trace`, it will be ignored.
    :returns:
        modes: The mode sizes of a grain size distribution.
        frequencies: The frequencies of corresponding mode sizes.
    """
    peaks, _ = find_peaks(distribution)
    position = classes[peaks] if is_geometric else classes_phi[peaks]
    values = distribution[peaks]
    not_trace = np.greater_equal(values, trace)
    return tuple(position[not_trace]), tuple(values[not_trace])


def to_cumulative(distribution: ndarray, expand=False) -> ndarray:
    """
    Get the cumulative distribution.

    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :param expand: If `True`, `0.0` and `1.0` will be added at the head and tail, respectively.
    :return: A numpy array contains the cumulative frequencies.
    """
    if expand:
        cumulative = np.zeros(len(distribution) + 2)
        cumulative[0] = 0.0
        cumulative[-1] = 1.0
        cumulative[1:-1] = np.cumsum(distribution)
    else:
        cumulative = np.cumsum(distribution)
    return cumulative


def reversed_phi_ppf(classes_phi: ndarray, distribution: ndarray) -> \
        Callable[[Union[int, float, ndarray]], Union[int, float, ndarray]]:
    """
    Get the reversed percent point function (PPF) of a grain size distribution.
    Because the grain size classes in phi values is isometric, it's more suitable for interpolation.
    And because coarser particles have smaller phi values, the PPF is reversed.

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :return: A scipy `interp1d` object which is callable and can be regarded as a PPF function.
    """
    interval = interval_phi(classes_phi)
    expand_classes = np.linspace(classes_phi[0] - interval, classes_phi[-1] + interval, len(classes_phi) + 2)
    cumulative = to_cumulative(distribution, expand=True)
    cumulative = np.array(cumulative)
    ppf = interp1d(cumulative, expand_classes, kind="slinear")
    return ppf


def interval_phi(classes_phi: ndarray) -> float:
    """
    Get the phi interval of grain size classes.

    :param classes_phi: The grain size classes in phi values.
    :return: The interval in phi values.
    """
    return abs((classes_phi[0] - classes_phi[-1]) / (len(classes_phi) - 1))


def to_phi(sizes: Union[int, float, ndarray]):
    """
    Convert the grain sizes from microns to phi values.

    :param sizes: The grain sizes in microns.
    :return: The grain sizes in phi values.
    """
    return -np.log2(sizes / 1000.0)


def to_microns(sizes_phi: ndarray):
    """
    Convert the grain sizes from phi values to microns.

    :param sizes_phi: The grain sizes in phi values.
    :return: The grain sizes in microns.
    """
    return np.exp2(-sizes_phi) * 1000.0


def cm(classes_phi: ndarray, distribution: ndarray) -> Tuple[float, float]:
    """
    Get the C and M values of a grain size distribution.
    C is the first percentile in phi scale, and M is the median in phi scale.

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :return: A tuple that contains the C and M values.
    """
    ppf = reversed_phi_ppf(classes_phi, distribution)
    CM = ppf(0.99), ppf(0.5)
    return CM


def ss(classes: ndarray, distribution: ndarray) -> Tuple[float, float]:
    """
        Get the proportion and mean size of sortable silt (10-63 microns).
        :param classes: The grain size classes in microns.
        :param distribution: The frequency distribution of grain size classes.
            Note, the sum of frequencies should be equal to 1.
        :return: A tuple that contains the proportion and mean size.
        """
    key = np.logical_and(classes > 10.0, classes < 63.0)
    proportion_ss = np.sum(distribution[key])
    mean_ss = np.exp(np.sum(distribution[key] * np.log(classes[key])) / proportion_ss)
    return proportion_ss, mean_ss


# The following five formulas of calculating the statistical parameters referred to Blott & Pye (2001)'s work
# DOI: 10.1002/esp.261
def arithmetic(classes: ndarray, distribution: ndarray) -> Dict[str, float]:
    """
    Calculate the basic statistical parameters.
    Follow the "arithmetic method of moments" in Blott & Pye (2001).

    :param classes: The grain size classes in microns.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :return: A `dict` that contains the basic statistical parameters.
        `dict(mean=..., std=..., skewness=..., kurtosis=...)`
    """
    mean = np.sum(classes * distribution)
    std = np.sqrt(np.sum(distribution * (classes - mean) ** 2))
    skewness = np.sum(distribution * (classes - mean) ** 3) / (std ** 3)
    kurtosis = np.sum(distribution * (classes - mean) ** 4) / (std ** 4)
    return dict(mean=mean, std=std, skewness=skewness, kurtosis=kurtosis)


def geometric(classes: ndarray, distribution: ndarray) -> Dict[str, Union[float, str]]:
    """
    Calculate the basic statistical parameters.
    Follow the "geometric method of moments" in Blott & Pye (2001).

    :param classes: The grain size classes in microns.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :return: A `dict` that contains the basic statistical parameters and corresponding descriptions.
        `dict(mean=..., std=..., skewness=..., kurtosis=..., std_description=..., skewness_description=...,
            kurtosis_description=...)`
    """

    def std_description(std_value):
        if std_value < 1.27:
            return "Very well sorted"
        elif std_value < 1.41:
            return "Well sorted"
        elif std_value < 1.62:
            return "Moderately well sorted"
        elif std_value < 2.00:
            return "Moderately sorted"
        elif std_value < 4.00:
            return "Poorly sorted"
        elif std_value < 16.00:
            return "Very poorly sorted"
        else:
            return "Extremely poorly sorted"

    def skewness_description(skewness_value):
        if skewness_value < -1.30:
            return "Very fine skewed"
        elif skewness_value < -0.43:
            return "Fine skewed"
        elif skewness_value < 0.43:
            return "Symmetrical"
        elif skewness_value < 1.30:
            return "Coarse skewed"
        else:
            return "Very coarse skewed"

    def kurtosis_description(kurtosis_value):
        if kurtosis_value < 1.70:
            return "Very platykurtic"
        elif kurtosis_value < 2.55:
            return "Platykurtic"
        elif kurtosis_value < 3.70:
            return "Mesokurtic"
        elif kurtosis_value < 7.40:
            return "Leptokurtic"
        else:
            return "Very leptokurtic"

    mean = np.exp(np.sum(distribution * np.log(classes)))
    std = np.exp(np.sqrt(np.sum(distribution * (np.log(classes) - np.log(mean)) ** 2)))
    skewness = np.sum(distribution * (np.log(classes) - np.log(mean)) ** 3) / (np.log(std) ** 3)
    kurtosis = np.sum(distribution * (np.log(classes) - np.log(mean)) ** 4) / (np.log(std) ** 4)

    return dict(mean=mean, std=std, skewness=skewness, kurtosis=kurtosis,
                std_description=std_description(std),
                skewness_description=skewness_description(skewness),
                kurtosis_description=kurtosis_description(kurtosis))


def logarithmic(classes_phi: ndarray, distribution: ndarray) -> Dict[str, Union[float, str]]:
    """
    Calculate the basic statistical parameters.
    Follow the "logarithmic method of moments" in Blott & Pye (2001).

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :return: A `dict` that contains the basic statistical parameters and corresponding descriptions.
        `dict(mean=..., std=..., skewness=..., kurtosis=..., std_description=..., skewness_description=...,
            kurtosis_description=...)`
    """

    def std_description(std_value):
        if std_value < 0.35:
            return "Very well sorted"
        elif std_value < 0.50:
            return "Well sorted"
        elif std_value < 0.70:
            return "Moderately well sorted"
        elif std_value < 1.00:
            return "Moderately sorted"
        elif std_value < 2.00:
            return "Poorly sorted"
        elif std_value < 4.00:
            return "Very poorly sorted"
        else:
            return "Extremely poorly sorted"

    def skewness_description(skewness_value):
        skewness_value = -skewness_value
        if skewness_value < -1.30:
            return "Very fine skewed"
        elif skewness_value < -0.43:
            return "Fine skewed"
        elif skewness_value < 0.43:
            return "Symmetrical"
        elif skewness_value < 1.30:
            return "Coarse skewed"
        else:
            return "Very coarse skewed"

    def kurtosis_description(kurtosis_value):
        if kurtosis_value < 1.70:
            return "Very platykurtic"
        elif kurtosis_value < 2.55:
            return "Platykurtic"
        elif kurtosis_value < 3.70:
            return "Mesokurtic"
        elif kurtosis_value < 7.40:
            return "Leptokurtic"
        else:
            return "Very leptokurtic"

    mean = np.sum(classes_phi * distribution)
    std = np.sqrt(np.sum(distribution * (classes_phi - mean) ** 2))
    skewness = np.sum(distribution * (classes_phi - mean) ** 3) / (std ** 3)
    kurtosis = np.sum(distribution * (classes_phi - mean) ** 4) / (std ** 4)

    return dict(mean=mean, std=std, skewness=skewness, kurtosis=kurtosis,
                std_description=std_description(std),
                skewness_description=skewness_description(skewness),
                kurtosis_description=kurtosis_description(kurtosis))


def logarithmic_fw57(_ppf: Callable[[Union[int, float, np.ndarray]], Union[int, float, np.ndarray]]) -> \
        Dict[str, Union[float, str]]:
    """
    Calculate the basic statistical parameters.
    Follow the "logarithmic (original) Folk & Ward (1957) graphical measures" in Blott & Pye (2001).

    :param _ppf: A `interp1d` object returned by `reversed_phi_ppf` function.
    :return: A `dict` that contains the basic statistical parameters and corresponding descriptions.
        `dict(mean=..., std=..., skewness=..., kurtosis=..., std_description=..., skewness_description=...,
            kurtosis_description=...)`
    """

    def ppf(x):
        if not isinstance(x, ndarray):
            x = np.array(x)
        return _ppf(1 - x)

    def std_description(std_value):
        if std_value < 0.35:
            return "Very well sorted"
        elif std_value < 0.50:
            return "Well sorted"
        elif std_value < 0.70:
            return "Moderately well sorted"
        elif std_value < 1.00:
            return "Moderately sorted"
        elif std_value < 2.00:
            return "Poorly sorted"
        elif std_value < 4.00:
            return "Very poorly sorted"
        else:
            return "Extremely poorly sorted"

    def skewness_description(skewness_value):
        skewness_value = -skewness_value
        if skewness_value < -0.3:
            return "Very fine skewed"
        elif skewness_value < -0.1:
            return "Fine skewed"
        elif skewness_value < 0.1:
            return "Symmetrical"
        elif skewness_value < 0.30:
            return "Coarse skewed"
        else:
            return "Very coarse skewed"

    def kurtosis_description(kurtosis_value):
        if kurtosis_value < 0.67:
            return "Very platykurtic"
        elif kurtosis_value < 0.90:
            return "Platykurtic"
        elif kurtosis_value < 1.11:
            return "Mesokurtic"
        elif kurtosis_value < 1.50:
            return "Leptokurtic"
        elif kurtosis_value < 3.00:
            return "Very leptokurtic"
        else:
            return "Extremely leptokurtic"

    mean = np.mean(ppf([0.16, 0.50, 0.84]))
    std = (ppf(0.84) - ppf(0.16)) / 4 + (ppf(0.95) - ppf(0.05)) / 6.6
    skewness = (ppf(0.16) + ppf(0.84) - 2 * ppf(0.50)) / 2 / (ppf(0.84) - ppf(0.16)) + (
            ppf(0.05) + ppf(0.95) - 2 * ppf(0.50)) / 2 / (ppf(0.95) - ppf(0.05))
    kurtosis = (ppf(0.95) - ppf(0.05)) / (2.44 * (ppf(0.75) - ppf(0.25)))

    return dict(mean=mean, std=std, skewness=skewness, kurtosis=kurtosis,
                std_description=std_description(std),
                skewness_description=skewness_description(skewness),
                kurtosis_description=kurtosis_description(kurtosis))


def geometric_fw57(_ppf: Callable[[Union[int, float, np.ndarray]], Union[int, float, np.ndarray]]) -> \
        Dict[str, Union[float, str]]:
    """
    Calculate the basic statistical parameters.
    Follow the "geometric (modified) Folk & Ward (1957) graphical measures" in Blott & Pye (2001).

    :param _ppf: A `interp1d` object returned by `reversed_phi_ppf` function.
    :return: A `dict` that contains the basic statistical parameters and corresponding descriptions.
        `dict(mean=..., std=..., skewness=..., kurtosis=..., std_description=..., skewness_description=...,
            kurtosis_description=...)`
    """

    def log_ppf(x):
        if not isinstance(x, ndarray):
            x = np.array(x)
        return np.log(to_microns(_ppf(x)))

    def std_description(std_value):
        if std_value < 1.27:
            return "Very well sorted"
        elif std_value < 1.41:
            return "Well sorted"
        elif std_value < 1.62:
            return "Moderately well sorted"
        elif std_value < 2.00:
            return "Moderately sorted"
        elif std_value < 4.00:
            return "Poorly sorted"
        elif std_value < 16.00:
            return "Very poorly sorted"
        else:
            return "Extremely poorly sorted"

    def skewness_description(skewness_value):
        if skewness_value < -0.3:
            return "Very fine skewed"
        elif skewness_value < -0.1:
            return "Fine skewed"
        elif skewness_value < 0.1:
            return "Symmetrical"
        elif skewness_value < 0.30:
            return "Coarse skewed"
        else:
            return "Very coarse skewed"

    def kurtosis_description(kurtosis_value):
        if kurtosis_value < 0.67:
            return "Very platykurtic"
        elif kurtosis_value < 0.90:
            return "Platykurtic"
        elif kurtosis_value < 1.11:
            return "Mesokurtic"
        elif kurtosis_value < 1.50:
            return "Leptokurtic"
        elif kurtosis_value < 3.00:
            return "Very leptokurtic"
        else:
            return "Extremely leptokurtic"

    mean = np.exp(np.mean(log_ppf([0.16, 0.50, 0.84])))
    std = np.exp((log_ppf(0.84) - log_ppf(0.16)) / 4 + (log_ppf(0.95) - log_ppf(0.05)) / 6.6)
    skewness = (log_ppf(0.16) + log_ppf(0.84) - 2 * log_ppf(0.50)) / 2 / (log_ppf(0.84) - log_ppf(0.16)) + (
            log_ppf(0.05) + log_ppf(0.95) - 2 * log_ppf(0.50)) / 2 / (log_ppf(0.95) - log_ppf(0.05))
    kurtosis = (log_ppf(0.95) - log_ppf(0.05)) / (2.44 * (log_ppf(0.75) - log_ppf(0.25)))

    return dict(mean=mean, std=std, skewness=skewness, kurtosis=kurtosis,
                std_description=std_description(std),
                skewness_description=skewness_description(skewness),
                kurtosis_description=kurtosis_description(kurtosis))


# Referred to Blott & Pye (2012)'s grain size scale
# DOI: 10.1111/j.1365-3091.2012.01335.x
def scale_description(phi: Union[int, float]) -> Tuple[str, str]:
    """
    Get the size description of a φ value. Follow Blott & Pye (2012)'s grain size scale.

    :param phi: The phi value of grain size.
    :returns:
        adj: The adjective of a grade, e.g., "Very large".
        grade: The name of this grade, e.g., "Boulder".
    """
    if phi < -11:
        return "", "Megaclasts"
    if phi < -10:
        return "Very large", "Boulder"
    elif phi < -9:
        return "Large", "Boulder"
    elif phi < -8:
        return "Medium", "Boulder"
    elif phi < -7:
        return "Small", "Boulder"
    elif phi < -6:
        return "Very small", "Boulder"
    elif phi < -5:
        return "Very coarse", "Gravel"
    elif phi < -4:
        return "Coarse", "Gravel"
    elif phi < -3:
        return "Medium", "Gravel"
    elif phi < -2:
        return "Fine", "Gravel"
    elif phi < -1:
        return "Very fine", "Gravel"
    elif phi < 0:
        return "Very coarse", "Sand"
    elif phi < 1:
        return "Coarse", "Sand"
    elif phi < 2:
        return "Medium", "Sand"
    elif phi < 3:
        return "Fine", "Sand"
    elif phi < 4:
        return "Very fine", "Sand"
    elif phi < 5:
        return "Very coarse", "Silt"
    elif phi < 6:
        return "Coarse", "Silt"
    elif phi < 7:
        return "Medium", "Silt"
    elif phi < 8:
        return "Fine", "Silt"
    elif phi < 9:
        return "Very fine", "Silt"
    elif phi < 10:
        return "Very coarse", "Clay"
    elif phi < 11:
        return "Coarse", "Clay"
    elif phi < 12:
        return "Medium", "Clay"
    elif phi < 13:
        return "Fine", "Clay"
    else:
        return "Very fine", "Clay"


def _all_scales():
    return tuple([scale_description(phi) for phi in np.linspace(-11.5, 13.5, 26)])


# GSM: Gravel, Sand, Mud
def proportions_gsm(classes_phi: ndarray, distribution: ndarray) -> \
        Tuple[float, float, float]:
    """
    Get the proportions `[0, 1]` of gravel, sand and mud.

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :returns:
        g: The proportion of gravel (-6 <= φ < -1).
        s: The proportion of sand (-1 <= φ < 4).
        m: The proportion of mud (φ >= 4).
    """
    gravel = np.max(np.sum(distribution[np.all([np.greater_equal(classes_phi, -6), np.less(classes_phi, -1)], axis=0)]))
    sand = np.max(np.sum(distribution[np.all([np.greater_equal(classes_phi, -1), np.less(classes_phi, 4)], axis=0)]))
    mud = np.max(np.sum(distribution[np.greater_equal(classes_phi, 4)]))
    return gravel, sand, mud


# SSC: Sand, Silt, Clay
def proportions_ssc(classes_phi: ndarray, distribution: ndarray) -> \
        Tuple[float, float, float]:
    """
    Get the proportions `[0, 1]` of sand, silt and clay.

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :returns:
        s: The proportion of sand (-1 <= φ < 4).
        si: The proportion of silt (4 <= φ < 9).
        c: The proportion of clay (φ >= 9).
    """
    sand = np.max(np.sum(distribution[np.all([np.greater_equal(classes_phi, -1), np.less(classes_phi, 4)], axis=0)]))
    silt = np.max(np.sum(distribution[np.all([np.greater_equal(classes_phi, 4), np.less(classes_phi, 9)], axis=0)]))
    clay = np.max(np.sum(distribution[np.greater_equal(classes_phi, 9)]))
    return sand, silt, clay


# BGSSC: Boulder, Gravel, Sand, Silt, Clay
def proportions_bgssc(classes_phi: ndarray, distribution: ndarray) -> \
        Tuple[float, float, float, float, float]:
    """
    Get the proportions `[0, 1]` of boulder, gravel, sand, silt and clay.

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :returns:
        b: The proportion of boulder (φ < -6).
        g: The proportion of gravel (-6 <= φ < -1).
        s: The proportion of sand (-1 <= φ < 4).
        si: The proportion of silt (4 <= φ < 9).
        c:  The proportion of clay (φ >= 9).
    """
    boulder = np.max(np.sum(distribution[np.less(classes_phi, -6)]))
    gravel = np.max(np.sum(distribution[np.all([np.greater_equal(classes_phi, -6), np.less(classes_phi, -1)], axis=0)]))
    sand = np.max(np.sum(distribution[np.all([np.greater_equal(classes_phi, -1), np.less(classes_phi, 4)], axis=0)]))
    silt = np.max(np.sum(distribution[np.all([np.greater_equal(classes_phi, 4), np.less(classes_phi, 9)], axis=0)]))
    clay = np.max(np.sum(distribution[np.greater_equal(classes_phi, 9)]))
    return boulder, gravel, sand, silt, clay


def all_proportions(classes_phi: ndarray, distribution: ndarray):
    """
    Get the proportions `[0, 1]` of all grades in `scale_description`.

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :return: The proportions of all grades.
    """
    proportions = dict()
    all_scales = _all_scales()
    for scale in all_scales:
        proportions[scale] = 0.0
    for phi, freq in zip(classes_phi, distribution):
        scale = scale_description(phi)
        proportions[scale] += freq
    return proportions


def group_gsm_folk54(gravel: float, sand: float, mud: float, trace=0.01):
    """
    Get the classification group of the gravel-sand-mud scheme of Folk (1954).

    :param gravel: The proportion of gravel (-6 <= φ < -1).
    :param sand: The proportion of sand (-1 <= φ < 4).
    :param mud: The proportion of mud (φ >= 4).
    :param trace: The `trace` is used to determine if it's `Slightly Gravelly`.
    :return: The name of this group.
    """
    if mud == 0.0:
        ratio = np.inf
    else:
        ratio = sand / mud

    if gravel < trace:
        if ratio < 1 / 9:
            return "Mud"
        elif ratio < 1:
            return "Sandy Mud"
        elif ratio < 9:
            return "Muddy Sand"
        else:
            return "Sand"
    elif gravel < 0.05:
        if ratio < 1 / 9:
            return "Slightly Gravelly Mud"
        elif ratio < 1:
            return "Slightly Gravelly Sandy Mud"
        elif ratio < 9:
            return "Slightly Gravelly Muddy Sand"
        else:
            return "Slightly Gravelly Sand"
    elif gravel < 0.3:
        if ratio < 1:
            return "Gravelly Mud"
        elif ratio < 9:
            return "Gravelly Muddy Sand"
        else:
            return "Gravelly Sand"
    elif gravel < 0.8:
        if ratio < 1:
            return "Muddy Gravel"
        elif ratio < 9:
            return "Muddy Sandy Gravel"
        else:
            return "Sandy Gravel"
    else:
        return "Gravel"


def group_ssc_folk54(sand: float, silt: float, clay: float):
    """
    Get the classification group of the sand-silt-clay scheme of Folk (1954).

    :param sand: The proportion of sand (-1 <= φ < 4).
    :param silt: The proportion of silt (4 <= φ < 9).
    :param clay: The proportion of clay (φ >= 9).
    :return: The name of this group.
    """
    if clay == 0.0:
        ratio = np.inf
    else:
        ratio = silt / clay
    if sand < 0.1:
        if ratio < 1 / 3:
            return "Clay"
        elif ratio < 2 / 3:
            return "Mud"
        else:
            return "Slit"
    elif sand < 0.5:
        if ratio < 1 / 3:
            return "Sandy Clay"
        elif ratio < 2 / 3:
            return "Sandy Mud"
        else:
            return "Sandy Slit"
    elif sand < 0.9:
        if ratio < 1 / 3:
            return "Clayey Sand"
        elif ratio < 2 / 3:
            return "Muddy Sand"
        else:
            return "Silty Sand"
    else:
        return "Sand"


def group_folk54(classes_phi: ndarray, distribution: ndarray):
    """
    Get the classification group of Folk (1954)'s scheme.

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :return: The name of this group.
    """
    gravel, sand, mud = proportions_gsm(classes_phi, distribution)
    if gravel > 0.0:
        return group_gsm_folk54(gravel, sand, mud)
    else:
        sand, silt, clay = proportions_ssc(classes_phi, distribution)
        return group_ssc_folk54(sand, silt, clay)


# map the symbol to its corresponding description
GROUP_BP12_SYMBOL_MAP = {
    "G": "Gravel",
    "S": "Sand",
    "SI": "Silt",
    "M": "Mud",
    "C": "Clay",
    "g": "Gravelly",
    "s": "Sandy",
    "si": "Silty",
    "m": "Muddy",
    "c": "Clayey",
    "(g)": "Slightly gravelly",
    "(s)": "Slightly sandy",
    "(si)": "Slightly silty",
    "(m)": "Slightly muddy",
    "(c)": "Slightly clayey",
    "(vg)": "Very slightly gravelly",
    "(vs)": "Very slightly sandy",
    "(vsi)": "Very slightly silty",
    "(vm)": "Very slightly muddy",
    "(vc)": "Very slightly clayey"}


def group_gsm_bp12(gravel: float, sand: float, mud: float, trace=0.01) \
        -> Tuple[List[str], List[str]]:
    """
    Get the classification group of the gravel-sand-mud scheme of Blott & Pye (2012).

    :param gravel: The proportion of gravel (-6 <= φ < -1).
    :param sand: The proportion of sand (-1 <= φ < 4).
    :param mud: The proportion of mud (φ >= 4).
    :param trace: The `trace` is used to determine if it's `Very slightly` or `Slightly`.
    :returns:
        symbols: The symbols of this group. Use `"".join(symbols)` to get the whole symbol.
        descriptions: The descriptions of this group.
            Use `string.capwords(" ".join(descriptions))` to get the complete description.
    """
    tags = {"vs": [], "s": [], "adj": [], "n": []}
    if gravel < trace:
        pass
    elif gravel < 0.05:
        tags["vs"].append("(vg)")
    elif gravel < 0.2:
        tags["s"].append("(g)")
    elif gravel < max(sand, mud):
        tags["adj"].append("g")
    else:
        tags["n"].append("G")

    if sand < trace:
        pass
    elif sand < 0.05:
        tags["vs"].append("(vs)")
    elif sand < 0.2:
        tags["s"].append("(s)")
    elif sand < max(gravel, mud):
        tags["adj"].append("s")
    else:
        tags["n"].append("S")

    if mud < trace:
        pass
    elif mud < 0.05:
        tags["vs"].append("(vm)")
    elif mud < 0.2:
        tags["s"].append("(m)")
    elif mud < max(gravel, sand):
        tags["adj"].append("m")
    else:
        tags["n"].append("M")

    symbols = tags["vs"] + tags["s"] + tags["adj"] + tags["n"]
    description = [GROUP_BP12_SYMBOL_MAP[s] for s in symbols]
    return symbols, description


def group_ssc_bp12(sand: float, silt: float, clay: float, trace=0.01) \
        -> Tuple[List[str], List[str]]:
    """
    Get the classification group of the sand-silt-clay scheme of Blott & Pye (2012).

    :param sand: The proportion of sand (-1 <= φ < 4).
    :param silt: The proportion of silt (4 <= φ < 9).
    :param clay: The proportion of clay (φ >= 9).
    :param trace: The `trace` is used to determine if it's `Very slightly` or `Slightly`.
    :returns:
        symbols: The symbols of this group. Use `"".join(symbols)` to get the whole symbol.
        descriptions: The descriptions of this group.
            Use `string.capwords(" ".join(descriptions))` to get the complete description.
    """
    tags = {"vs": [], "s": [], "adj": [], "n": []}
    if sand < trace:
        pass
    elif sand < 0.05:
        tags["vs"].append("(vs)")
    elif sand < 0.2:
        tags["s"].append("(s)")
    elif sand < max(sand, clay):
        tags["adj"].append("s")
    else:
        tags["n"].append("S")

    if silt < trace:
        pass
    elif silt < 0.05:
        tags["vs"].append("(vsi)")
    elif silt < 0.2:
        tags["s"].append("(si)")
    elif silt < max(sand, clay):
        tags["adj"].append("si")
    else:
        tags["n"].append("SI")

    if clay < trace:
        pass
    elif clay < 0.05:
        tags["vs"].append("(vc)")
    elif clay < 0.2:
        tags["s"].append("(c)")
    elif clay < max(sand, silt):
        tags["adj"].append("c")
    else:
        tags["n"].append("C")

    symbols = tags["vs"] + tags["s"] + tags["adj"] + tags["n"]
    description = [GROUP_BP12_SYMBOL_MAP[s] for s in symbols]
    return symbols, description


def group_bp12(classes_phi: ndarray, distribution: ndarray) \
        -> Tuple[List[str], List[str]]:
    """
    Get the classification group of Blott & Pye (2012)'s scheme.

    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :returns:
        symbols: The symbols of this group. Use `"".join(symbols)` to get the whole symbol.
        descriptions: The descriptions of this group.
            Use `string.capwords(" ".join(descriptions))` to get the complete description.
    """
    gravel, sand, mud = proportions_gsm(classes_phi, distribution)
    if gravel > 0.0:
        return group_gsm_bp12(gravel, sand, mud)
    else:
        sand, silt, clay = proportions_ssc(classes_phi, distribution)
        return group_ssc_bp12(sand, silt, clay)


def major_statistics(classes: ndarray, classes_phi: ndarray,
                     distribution: ndarray, is_geometric=False, is_fw57=False) -> dict:
    """
    Get the statistical parameters of a grain size distribution.

    :param classes: The grain size classes in microns.
    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :param is_geometric: If `True`, the grain sizes are in microns, else they are in phi values.
    :param is_fw57: If `True`, it will use Folk & Ward (1957)'s graphical method.
        Else, it will use the method of statistical moments.
    :return: The `dict` contains these statistical parameters, including
        `mean`, `std`, `skewness`, `kurtosis`,
        `mode`, `modes`, `median`, `mean_description`, `std_description`,
        `skewness_description`, `kurtosis_description`.
    """
    reverse_phi_ppf = reversed_phi_ppf(classes_phi, distribution)
    median_phi = np.max(reverse_phi_ppf(0.5))
    first_percentile_phi = np.max(reverse_phi_ppf(0.99))
    proportion_ss, mean_ss = ss(classes, distribution)
    statistics = {}
    if is_geometric:
        if is_fw57:
            statistics.update(geometric_fw57(reverse_phi_ppf))
        else:
            statistics.update(geometric(classes, distribution))
        mean_phi = to_phi(statistics["mean"])
        statistics["median"] = to_microns(median_phi)
        statistics["first_percentile"] = to_microns(first_percentile_phi)
        statistics["proportion_ss"] = proportion_ss
        statistics["mean_ss"] = mean_ss
    else:
        if is_fw57:
            statistics = logarithmic_fw57(reverse_phi_ppf)
        else:
            statistics = logarithmic(classes_phi, distribution)
        mean_phi = statistics["mean"]
        statistics["median"] = median_phi
        statistics["first_percentile"] = first_percentile_phi
        statistics["proportion_ss"] = proportion_ss
        statistics["mean_ss"] = to_phi(mean_ss)
    mean_description = string.capwords(" ".join(scale_description(mean_phi))).strip()
    statistics["mean_description"] = mean_description
    statistics["mode"] = mode(classes, classes_phi, distribution, is_geometric=is_geometric)
    statistics["modes"] = modes(classes, classes_phi, distribution, is_geometric=is_geometric)[0]
    return statistics


def all_statistics(classes: ndarray, classes_phi: ndarray, distribution: ndarray):
    """
    Get the statistical parameters and classification groups of all methods of a grain size distribution.

    :param classes: The grain size classes in microns.
    :param classes_phi: The grain size classes in phi values.
    :param distribution: The frequency distribution of grain size classes.
        Note, the sum of frequencies should be equal to 1.
    :return: The `dict` contains the statistical parameters of all methods, including
        `arithmetic`, `geometric`, `logarithmic`, `geometric_fw57`, `logarithmic_fw57`;
        and the proportions of different grades, including `proportions_gsm`, `proportions_ssc`,
        `proportions_bgssc`, `proportions`; and the classification groups,
        including `group_folk54` and `group_bp12`.
    """
    result = {
        "arithmetic": arithmetic(classes, distribution),
        "geometric": major_statistics(classes, classes_phi, distribution, is_geometric=True, is_fw57=False),
        "logarithmic": major_statistics(classes, classes_phi, distribution, is_geometric=False, is_fw57=False),
        "geometric_fw57": major_statistics(classes, classes_phi, distribution, is_geometric=True, is_fw57=True),
        "logarithmic_fw57": major_statistics(classes, classes_phi, distribution, is_geometric=False, is_fw57=True),
        "proportions_gsm": proportions_gsm(classes_phi, distribution),
        "proportions_ssc": proportions_ssc(classes_phi, distribution),
        "proportions_bgssc": proportions_bgssc(classes_phi, distribution),
        "proportions": all_proportions(classes_phi, distribution),
        "group_folk54": group_folk54(classes_phi, distribution)}

    bp12_symbols, bp12_descriptions = group_bp12(classes_phi, distribution)
    bp12_symbol = "".join(bp12_symbols)
    bp12_description = string.capwords(" ".join(bp12_descriptions))
    result["_group_bp12_symbols"] = bp12_symbols
    result["group_bp12_symbol"] = bp12_symbol
    result["group_bp12"] = bp12_description
    return result



__all__ = ["try_udm"]

import logging
import time
from typing import *

import numpy as np
import torch
from numpy import ndarray

from .models import KernelType, Dataset, UDMResult, ArtificialDataset
from .kernels import ProportionModule, get_kernel

torch.set_default_dtype(torch.float64)


class UDMModule(torch.nn.Module):
    def __init__(self, n_samples: int, n_components: int, classes_phi: np.ndarray,
                 kernel_type: KernelType, x0: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_components = n_components
        self.n_classes = len(classes_phi)
        self._interval_phi = np.abs((classes_phi[0] - classes_phi[-1]) / (classes_phi.shape[0] - 1))
        self._classes_phi = torch.nn.Parameter(torch.from_numpy(classes_phi).repeat(n_samples, n_components, 1),
                                               requires_grad=False)
        self.kernel_type = kernel_type
        self.proportions = ProportionModule(n_samples, n_components)
        self.components = get_kernel(kernel_type, n_samples, self.n_components, self.n_classes, x0)

    def forward(self) -> Tuple[torch.Tensor, torch.Tensor]:
        # n_samples x 1 x n_members
        proportions = self.proportions()
        # n_samples x n_members x n_classes
        components = self.components(self._classes_phi, self._interval_phi)
        return proportions, components

    @property
    def all_parameters(self) -> np.ndarray:
        with torch.no_grad():
            all_parameters = torch.cat([self.components._params, self.proportions._params], dim=1)
            all_parameters = all_parameters.detach().cpu().numpy()
        return all_parameters


def try_udm(dataset: Union[ArtificialDataset, Dataset], kernel_type: KernelType, n_components: int, x0: ndarray = None,
            device="cpu", pretrain_epochs=200, min_epochs=200, max_epochs=2000, precision: Union[int, float] = 6,
            learning_rate=5e-3, betas=(0.8, 0.5), constraint_level: Union[int, float] = 2.0, need_history=True,
            logger: logging.Logger = None, progress_callback: Callable[[float], None] = None) -> UDMResult:
    assert isinstance(dataset, (ArtificialDataset, Dataset))
    assert isinstance(kernel_type, KernelType)
    assert isinstance(n_components, int)
    if x0 is not None:
        assert isinstance(x0, ndarray)
        assert x0.ndim == 2
        assert x0.shape[1] == n_components
        x0 = x0.astype(np.float64)
    available_devices = ["cpu"]
    if torch.cuda.is_available():
        available_devices.append("cuda")
    available_devices.extend([f"cuda:{i}" for i in range(torch.cuda.device_count())])
    if "cuda:0" in available_devices:
        available_devices.remove("cuda:0")
    assert device in available_devices
    assert isinstance(pretrain_epochs, int)
    assert isinstance(min_epochs, int)
    assert isinstance(max_epochs, int)
    assert isinstance(precision, (int, float))
    assert isinstance(learning_rate, float)
    assert isinstance(betas, tuple)
    assert len(betas) == 2
    beta1, beta2 = betas
    assert isinstance(beta1, float)
    assert isinstance(beta2, float)
    assert isinstance(constraint_level, (int, float))
    assert pretrain_epochs >= 0
    assert min_epochs > 0
    assert max_epochs > 0
    assert 1.0 < precision < 100.0
    assert learning_rate > 0.0
    assert 0.0 < beta1 < 1.0
    assert 0.0 < beta2 < 1.0
    assert -10.0 < constraint_level < 10.0
    if logger is None:
        logger = logging.getLogger("QGrain")
    else:
        assert isinstance(logger, logging.Logger)

    start_text = f"""Performing the UDM algorithm on the dataset ({dataset.name}).
    Kernel type: {kernel_type.name}
    Number of end members: {n_components}
    x0: {x0 if x0 is None else x0.tolist()}
    Device: {device}
    Pretrain epochs: {pretrain_epochs}
    Minimum epochs: {min_epochs}
    Maximum epochs: {max_epochs}
    Precision decimals (10^-x): {precision}
    Learning rate: {learning_rate}
    Betas of weight decay: {betas}
    Constraint level: {constraint_level}
    Need history: {need_history}"""
    logger.debug(start_text)

    observation = torch.from_numpy(dataset.distributions.astype(np.float64)).to(device)
    udm = UDMModule(len(dataset), n_components, dataset.classes_phi.astype(np.float64), kernel_type, x0).to(device)
    optimizer = torch.optim.Adam(udm.parameters(), lr=learning_rate, betas=betas)
    distribution_loss_series = []
    component_loss_series = []
    history: List[ndarray] = [udm.all_parameters]
    udm.components.requires_grad_(False)
    max_total_epochs = pretrain_epochs + max_epochs
    start = time.time()
    for pretrain_epoch in range(pretrain_epochs):
        proportions, components = udm()
        prediction = (proportions @ components).squeeze(1)
        distribution_loss = torch.log10(torch.mean(torch.square(observation - prediction)))
        distribution_loss_series.append(distribution_loss.item())
        component_loss_series.append(0.0)
        optimizer.zero_grad()
        distribution_loss.backward()
        torch.nn.utils.clip_grad_norm_(udm.parameters(), 1e-1)
        optimizer.step()
        if need_history:
            history.append(udm.all_parameters)
        if progress_callback is not None:
            progress_callback(pretrain_epoch / max_total_epochs)

    udm.components.requires_grad_(True)
    for epoch in range(max_epochs):
        # train
        proportions, components = udm()
        prediction = (proportions @ components).squeeze(1)
        distribution_loss = torch.log10(torch.mean(torch.square(prediction - observation)))
        # component_loss = torch.log10(torch.mean(torch.std(components, dim=0)))
        component_loss = torch.mean(torch.std(components, dim=0))
        loss = distribution_loss + (10 ** constraint_level) * component_loss
        if np.isnan(loss.item()):
            logger.warning("Loss is NaN, training has beem terminated.")
            break
        distribution_loss_series.append(distribution_loss.item())
        component_loss_series.append((10 ** constraint_level) * component_loss.item())
        optimizer.zero_grad()
        loss.backward()
        torch.nn.utils.clip_grad_norm_(udm.parameters(), 1e-1)
        optimizer.step()
        if need_history:
            history.append(udm.all_parameters)
        if progress_callback is not None:
            progress_callback((pretrain_epochs + epoch) / max_total_epochs)
        if epoch > min_epochs:
            delta_loss = np.mean(distribution_loss_series[-100:-80]) - np.mean(distribution_loss_series[-20:])
            if delta_loss < 10 ** (-precision):
                break

    # algorithm finished, preparing the result
    if device[:4] == "cuda":
        torch.cuda.synchronize()
    time_spent = time.time() - start
    settings = dict(device=device, pretrain_epochs=pretrain_epochs, min_epochs=min_epochs,
                    max_epochs=max_epochs, precision=precision, learning_rate=learning_rate,
                    betas=betas, constraint_level=constraint_level, need_history=need_history)
    distribution_loss_series = np.array(distribution_loss_series)
    component_loss_series = np.array(component_loss_series)
    loss_series = {"total": distribution_loss_series + component_loss_series,
                   "distribution": distribution_loss_series,
                   "component": component_loss_series}
    if need_history:
        parameters = np.concatenate([np.expand_dims(p, axis=0) for p in history], axis=0)
    else:
        parameters = np.expand_dims(udm.all_parameters, axis=0)
    result = UDMResult(dataset, kernel_type, n_components, parameters, time_spent, x0, settings, loss_series)
    if progress_callback is not None:
        progress_callback(1.0)
    return result



__all__ = ["UDMAnalyzer"]

import logging
import pickle
from typing import *

import numpy as np
from PySide6 import QtCore, QtWidgets
from grpc._channel import _InactiveRpcError

from ..models import Dataset, KernelType, UDMResult
from ..charts.UDMResultChart import UDMResultChart
from ..protos.client import QGrainClient
from ..io import save_udm
from .UDMSettings import UDMSettings
from .ParameterEditor import ParameterEditor


class UDMAnalyzer(QtWidgets.QWidget):
    logger = logging.getLogger("QGrain.UDMAnalyzer")
    SUPPORT_KERNELS = (
        (KernelType.Normal, "Normal"),
        (KernelType.SkewNormal, "Skew Normal"),
        (KernelType.Weibull, "Weibull"),
        (KernelType.GeneralWeibull, "General Weibull"))

    def __init__(self, setting_dialog: UDMSettings, parameter_editor: ParameterEditor,
                 client: QGrainClient = None, parent: QtWidgets.QWidget = None):
        super().__init__(parent=parent)
        assert setting_dialog is not None
        assert parameter_editor is not None
        self.setting_dialog = setting_dialog
        self.parameter_editor = parameter_editor
        self._client = client
        self.setWindowTitle(self.tr("UDM Analyzer"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.control_group = QtWidgets.QGroupBox(self.tr("Control"))
        self.control_group.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.control_layout = QtWidgets.QGridLayout(self.control_group)
        self.kernel_label = QtWidgets.QLabel(self.tr("Kernel Type"))
        self.kernel_combo_box = QtWidgets.QComboBox()
        self.kernel_combo_box.addItems([name for _, name in self.SUPPORT_KERNELS])
        self.kernel_combo_box.setCurrentIndex(0)
        self.control_layout.addWidget(self.kernel_label, 0, 0)
        self.control_layout.addWidget(self.kernel_combo_box, 0, 1)
        self.n_components_label = QtWidgets.QLabel(self.tr("Number of Components"))
        self.n_members_input = QtWidgets.QSpinBox()
        self.n_members_input.setRange(1, 12)
        self.n_members_input.setValue(3)
        self.control_layout.addWidget(self.n_components_label, 1, 0)
        self.control_layout.addWidget(self.n_members_input, 1, 1)
        self.try_fit_button = QtWidgets.QPushButton(self.tr("Try Fit"))
        self.try_fit_button.clicked.connect(self.on_try_fit_clicked)
        self.try_fit_button.setEnabled(False)
        self.edit_parameter_button = QtWidgets.QPushButton(self.tr("Edit Parameters"))
        self.edit_parameter_button.clicked.connect(self.on_edit_parameter_clicked)
        self.edit_parameter_button.setEnabled(False)
        self.control_layout.addWidget(self.try_fit_button, 2, 0)
        self.control_layout.addWidget(self.edit_parameter_button, 2, 1)
        self.result_group = QtWidgets.QGroupBox(self.tr("Result"))
        self.result_group.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.result_layout = QtWidgets.QGridLayout(self.result_group)
        self.result_list_widget = QtWidgets.QListWidget()
        self.result_layout.addWidget(self.result_list_widget, 0, 0, 1, 2)
        self.remove_button = QtWidgets.QPushButton(self.tr("Remove"))
        self.remove_button.clicked.connect(self.on_remove_clicked)
        self.show_button = QtWidgets.QPushButton(self.tr("Show"))
        self.show_button.clicked.connect(self.on_show_clicked)
        self.remove_button.setEnabled(False)
        self.show_button.setEnabled(False)
        self.result_layout.addWidget(self.remove_button, 1, 0)
        self.result_layout.addWidget(self.show_button, 1, 1)
        self.chart_group = QtWidgets.QGroupBox(self.tr("Chart"))
        self.chart_group.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.chart_layout = QtWidgets.QGridLayout(self.chart_group)
        self.result_chart = UDMResultChart()
        self.chart_layout.addWidget(self.result_chart)
        self.splitter_1 = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.splitter_1.addWidget(self.control_group)
        self.splitter_1.addWidget(self.result_group)
        self.splitter_2 = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        self.splitter_2.addWidget(self.splitter_1)
        self.splitter_2.addWidget(self.chart_group)
        self.main_layout.addWidget(self.splitter_2)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self._dataset: Optional[Dataset] = None
        self._results: List[UDMResult] = []

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    @property
    def kernel_type(self) -> KernelType:
        kernel_type = self.SUPPORT_KERNELS[self.kernel_combo_box.currentIndex()][0]
        return kernel_type

    @property
    def n_members(self):
        return self.n_members_input.value()

    @property
    def n_results(self) -> int:
        return len(self._results)

    @property
    def selected_index(self):
        indexes = self.result_list_widget.selectedIndexes()
        if len(indexes) == 0:
            return 0
        else:
            return indexes[0].row()

    @property
    def selected_result(self):
        if self.n_results > 0:
            return self._results[self.selected_index]

    def on_dataset_loaded(self, dataset: Dataset):
        self._dataset = dataset
        self.try_fit_button.setEnabled(True)
        self.edit_parameter_button.setEnabled(True)

    def on_edit_parameter_clicked(self):
        if self._dataset is not None:
            target = np.mean(self._dataset.distributions, axis=0)
            target = target / np.sum(target)
            self.parameter_editor.setup_target(self._dataset.classes, target)
        self.parameter_editor.show()

    def on_try_fit_clicked(self):
        if self._dataset is None:
            self.logger.error("The dataset has not been loaded.")
            self.show_error(self.tr("The dataset has not been loaded."))
            return

        settings = {**self.setting_dialog.settings}
        if self.parameter_editor.parameter_enabled:
            self.logger.info("The parameters in Parameter Editor are enabled. They will be used preferentially!")
            settings["kernel_type"] = KernelType.__members__[self.parameter_editor.distribution_type.name]
            settings["n_components"] = self.parameter_editor.n_components
            settings["x0"] = self.parameter_editor.parameters[:-1, :].astype(np.float64)
        else:
            settings["kernel_type"] = self.kernel_type
            settings["n_components"] = self.n_members
        self.logger.debug(f"Try to perform the UDM algorithm. Algorithm settings: {settings}.")

        if self._client is not None and self._client.has_target:
            try:
                server_state = self._client.get_service_state()
                if len(self._dataset) > server_state["max_dataset_size"]:
                    self.logger.error(
                        f"The dataset size ({len(self._dataset)} samples) exceeded the limitation of remote grpc server.")
                    self.show_error(self.tr("The dataset size exceeded the limitation of remote grpc server."))
                    return
                # TODO: async & progress report
                result_or_msg = self._client.get_udm_result(self._dataset, **settings)
                if isinstance(result_or_msg, UDMResult):
                    self.add_results([result_or_msg])
                    self.result_chart.show_result(result_or_msg)
                    return
                else:
                    self.logger.error(f"The UDM fitting task failed: {result_or_msg}.")
                    self.show_error(self.tr("The UDM fitting task failed, please check the logs for more details."))
                    return
            except _InactiveRpcError:
                    self.logger.warning("The remote grpc server is not available.")

        self.try_fit_button.setEnabled(False)
        self.edit_parameter_button.setEnabled(False)
        progress_dialog = QtWidgets.QProgressDialog(
            self.tr("Performing the UDM algorithm..."), self.tr("Cancel"),
            0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

        def callback(progress: float):
            if progress_dialog.wasCanceled():
                raise StopIteration()
            progress_dialog.setValue(int(progress * 100))
            QtCore.QCoreApplication.processEvents()
        try:
            from ..udm import try_udm
            result = try_udm(self._dataset, **settings, progress_callback=callback)
            self.add_results([result])
            self.result_chart.show_result(result)
        except StopIteration:
            self.logger.info("The performing task was canceled.")
        finally:
            progress_dialog.close()
            self.try_fit_button.setEnabled(True)
            self.edit_parameter_button.setEnabled(True)

    @classmethod
    def get_result_name(cls, result: UDMResult):
        return f"{result.dataset.name} ({result.n_components}, {result.kernel_type.name})"

    def add_results(self, results: List[UDMResult]):
        if self.n_results == 0:
            self.remove_button.setEnabled(True)
            self.show_button.setEnabled(True)

        self._results.extend(results)
        self.result_list_widget.addItems([self.get_result_name(result) for result in results])

    def on_remove_clicked(self):
        if self.n_results == 0:
            return
        else:
            index = self.selected_index
            self._results.pop(index)
            self.result_list_widget.takeItem(index)

        if self.n_results == 0:
            self.remove_button.setEnabled(False)
            self.show_button.setEnabled(False)

    def on_show_clicked(self):
        result = self.selected_result
        if result is not None:
            self.result_chart.show_result(result)

    def on_show_animation_clicked(self):
        result = self.selected_result
        if result is not None:
            self.result_chart.show_animation(result)

    def load_result(self):
        filename, _  = self.file_dialog.getOpenFileName(
            self, self.tr("Choose the file which stores the dumped UDM result"),
            ".", "Dumped UDM Result (*.udm)")
        if filename is None or filename == "":
            return
        with open(filename, "rb") as f:
            result = pickle.load(f)
            if isinstance(result, UDMResult):
                self.add_results([result])
                self.logger.info(f"The dumped UDM result has been loaded.")
            else:
                self.logger.error("The binary file is invalid (i.e., the object in it is not the UDM result).")
                self.show_error(self.tr("The binary file is invalid (i.e., the object in it is not the UDM result)."))
                return

    def save_selected_result(self):
        if self.n_results == 0:
            self.show_error(self.tr("There is no UDM result."))
            return
        filename, _ = self.file_dialog.getSaveFileName(
            self, self.tr("Choose a filename to save the selected UDM result"),
            ".", "Microsoft Excel (*.xlsx);;Dumped UDM Result (*.udm)")
        if filename is None or filename == "":
            return
        # Excel
        if filename[-4:] == "xlsx":
            result = self.selected_result
            progress_dialog = QtWidgets.QProgressDialog(
                self.tr("Saving the UDM result..."), self.tr("Cancel"),
                0, 100, self)
            progress_dialog.setWindowTitle("QGrain")
            progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

            def callback(progress: float):
                if progress_dialog.wasCanceled():
                    raise StopIteration()
                progress_dialog.setValue(int(progress*100))
                QtCore.QCoreApplication.processEvents()
            try:
                save_udm(result, filename, progress_callback=callback, logger=self.logger)
            except StopIteration as e:
                self.logger.info("The saving task was canceled.")
            finally:
                progress_dialog.close()
        # Binary File
        else:
            with open(filename, "wb") as f:
                pickle.dump(self.selected_result, f)
                self.logger.info("The selected UDM result has been dumped.")

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("UDM Analyzer"))
        self.control_group.setTitle(self.tr("Control"))
        self.kernel_label.setText(self.tr("Kernel Type"))
        self.n_components_label.setText(self.tr("Number of Components"))
        self.try_fit_button.setText(self.tr("Try Fit"))
        self.edit_parameter_button.setText(self.tr("Edit Parameters"))
        self.result_group.setTitle(self.tr("Result"))
        self.remove_button.setText(self.tr("Remove"))
        self.show_button.setText(self.tr("Show"))
        self.chart_group.setTitle(self.tr("Chart"))



__all__ = ["ParameterTable"]


from PySide6 import QtCore, QtWidgets

from ..models import DistributionType, SSUResult


class ParameterTable(QtWidgets.QDialog):
    parameter_name_map = {
        DistributionType.Normal: ("Location", "Scale", "Weight"),
        DistributionType.SkewNormal: ("Shape", "Location", "Scale", "Weight"),
        DistributionType.Weibull: ("Shape", "Scale", "Weight"),
        DistributionType.GeneralWeibull: ("Shape", "Location", "Scale", "Weight")}

    def __init__(self, result: SSUResult, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        # make sure that the pyside2-lupdate.exe can recognize these keywords
        _ = [self.tr("Shape"), self.tr("Location"), self.tr("Scale"), self.tr("Weight")]
        self.main_layout = QtWidgets.QGridLayout(self)
        self.setMinimumSize(200, 160)
        self.setWindowTitle(self.tr("Resolved Parameters"))
        parameter_names = self.parameter_name_map[result.distribution_type]
        self.headers = [] # type: list[tuple[str, QtWidgets.QLabel]]
        for i, name in enumerate(parameter_names):
            header = QtWidgets.QLabel(self.tr(name))
            self.main_layout.addWidget(header, 0, i+1)
            self.headers.append((name, header))
        for i in range(len(result)):
            row = i+1
            component_name = QtWidgets.QLabel(f"C{i+1}")
            self.main_layout.addWidget(component_name, row, 0)
            for j in range(len(parameter_names)):
                col = j+1
                parameter = QtWidgets.QLabel(f"{result.parameters[-1, j, i]:.4f}")
                self.main_layout.addWidget(parameter, row, col)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("Resolved Parameters"))
        for name, header in self.headers:
            header.setText(self.tr(name))



__all__ = ["SSUAnalyzer"]

from typing import *

from PySide6 import QtCore, QtWidgets

from ..models import DistributionType, Dataset, SSUResult
from ..charts.DistributionChart import DistributionChart
from .ParameterEditor import ParameterEditor
from .SSUResultViewer import SSUResultViewer
from .SSUSettings import SSUSettings
from ..ssu import try_ssu


class SSUAnalyzer(QtWidgets.QWidget):
    SUPPORT_DISTRIBUTIONS = (
        (DistributionType.Normal, "Normal"),
        (DistributionType.SkewNormal, "Skew Normal"),
        (DistributionType.Weibull, "Weibull"),
        (DistributionType.GeneralWeibull, "General Weibull"))

    def __init__(self, setting_dialog: SSUSettings, parameter_editor: ParameterEditor, parent=None):
        super().__init__(parent=parent)
        assert setting_dialog is not None
        assert parameter_editor is not None
        self.setting_dialog = setting_dialog
        self.parameter_editor = parameter_editor
        self.setWindowTitle(self.tr("SSU Analyzer"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.control_group = QtWidgets.QGroupBox(self.tr("Control"))
        self.control_layout = QtWidgets.QGridLayout(self.control_group)
        self.distribution_label = QtWidgets.QLabel(self.tr("Distribution Type"))
        self.distribution_combo_box = QtWidgets.QComboBox()
        self.distribution_combo_box.addItems([name for _, name in self.SUPPORT_DISTRIBUTIONS])
        self.component_number_label = QtWidgets.QLabel(self.tr("Number of Components"))
        self.n_components_input = QtWidgets.QSpinBox()
        self.n_components_input.setRange(1, 10)
        self.n_components_input.setValue(3)
        self.control_layout.addWidget(self.distribution_label, 0, 0)
        self.control_layout.addWidget(self.distribution_combo_box, 0, 1)
        self.control_layout.addWidget(self.component_number_label, 1, 0)
        self.control_layout.addWidget(self.n_components_input, 1, 1)
        self.sample_index_label = QtWidgets.QLabel(self.tr("Sample Index"))
        self.sample_index_input = QtWidgets.QSpinBox()
        self.sample_index_input.valueChanged.connect(self.on_sample_index_changed)
        self.sample_index_input.setEnabled(False)
        self.control_layout.addWidget(self.sample_index_label, 2, 0)
        self.control_layout.addWidget(self.sample_index_input, 2, 1)
        self.sample_name_label = QtWidgets.QLabel(self.tr("Sample Name"))
        self.sample_name_display = QtWidgets.QLabel(self.tr("Unknown"))
        self.control_layout.addWidget(self.sample_name_label, 3, 0)
        self.control_layout.addWidget(self.sample_name_display, 3, 1)
        self.try_fit_button = QtWidgets.QPushButton(self.tr("Try Fit"))
        self.try_fit_button.setEnabled(False)
        self.try_fit_button.clicked.connect(self.on_try_fit_clicked)
        self.edit_parameter_button = QtWidgets.QPushButton(self.tr("Edit Parameters"))
        self.edit_parameter_button.clicked.connect(self.on_edit_parameter_clicked)
        self.control_layout.addWidget(self.try_fit_button, 4, 0)
        self.control_layout.addWidget(self.edit_parameter_button, 4, 1)
        self.try_previous_button = QtWidgets.QPushButton(self.tr("Try Previous"))
        self.try_previous_button.setEnabled(False)
        self.try_previous_button.clicked.connect(self.on_try_previous_clicked)
        self.try_next_button = QtWidgets.QPushButton(self.tr("Try Next"))
        self.try_next_button.setEnabled(False)
        self.try_next_button.clicked.connect(self.on_try_next_clicked)
        self.control_layout.addWidget(self.try_previous_button, 5, 0)
        self.control_layout.addWidget(self.try_next_button, 5, 1)
        self.chart_group = QtWidgets.QGroupBox(self.tr("Chart"))
        self.chart_layout = QtWidgets.QGridLayout(self.chart_group)
        self.result_chart = DistributionChart()
        self.chart_layout.addWidget(self.result_chart, 0, 0)
        self.result_group = QtWidgets.QGroupBox(self.tr("Result"))
        self.result_layout = QtWidgets.QGridLayout(self.result_group)
        self.result_view = SSUResultViewer()
        self.result_view.result_displayed.connect(self.on_result_displayed)
        self.result_view.result_referred.connect(self.parameter_editor.refer_result)
        self.result_layout.addWidget(self.result_view)
        self.splitter_1 = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.splitter_1.addWidget(self.control_group)
        self.splitter_1.addWidget(self.chart_group)
        self.splitter_2 = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        self.splitter_2.addWidget(self.splitter_1)
        self.splitter_2.addWidget(self.result_group)
        self.splitter_2.setStretchFactor(0, 1)
        self.splitter_2.setStretchFactor(1, 2)
        self.main_layout.addWidget(self.splitter_2, 0, 0)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self._dataset = None

    @property
    def distribution_type(self) -> DistributionType:
        distribution_type = self.SUPPORT_DISTRIBUTIONS[self.distribution_combo_box.currentIndex()][0]
        return distribution_type

    @property
    def n_components(self) -> int:
        return self.n_components_input.value()

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_dataset_loaded(self, dataset: Dataset):
        self._dataset = dataset
        self.sample_index_input.setRange(1, len(dataset))
        self.sample_index_input.setValue(1)
        self.on_sample_index_changed(1)
        self.sample_index_input.setEnabled(True)
        self.try_fit_button.setEnabled(True)
        self.edit_parameter_button.setEnabled(True)
        self.try_previous_button.setEnabled(True)
        self.try_next_button.setEnabled(True)

    def on_sample_index_changed(self, index):
        self.sample_name_display.setText(self._dataset[index - 1].name)

    def get_task(self, sample_index: int) -> Dict[str, Any]:
        task = {}
        sample = self._dataset[sample_index]
        task["sample"] = sample
        if self.parameter_editor.parameter_enabled:
            task["distribution_type"] = self.parameter_editor.distribution_type
            task["n_components"] = self.parameter_editor.n_components
            task["x0"] = self.parameter_editor.parameters
        else:
            task["distribution_type"] = self.distribution_type
            task["n_components"] = self.n_components
        task.update(self.setting_dialog.settings)
        return task

    def get_all_tasks(self) -> List[Dict[str, Any]]:
        tasks = []
        for i in range(len(self._dataset)):
            task = self.get_task(i)
            tasks.append(task)
        return tasks

    def on_result_displayed(self, result: SSUResult):
        self.result_chart.show_result(result)

    def do_test(self):
        self.try_fit_button.setEnabled(False)
        self.edit_parameter_button.setEnabled(False)
        self.try_previous_button.setEnabled(False)
        self.try_next_button.setEnabled(False)
        task = self.get_task(self.sample_index_input.value()-1)
        result, msg = try_ssu(**task)
        if isinstance(result, SSUResult):
            self.result_chart.show_result(result)
            self.result_view.add_result(result)
        else:
            self.show_error(self.tr("The SSU fitting task failed: {0}.").format(msg))
        self.try_fit_button.setEnabled(True)
        self.edit_parameter_button.setEnabled(True)
        self.try_previous_button.setEnabled(True)
        self.try_next_button.setEnabled(True)

    def on_edit_parameter_clicked(self):
        if self._dataset is not None:
            sample_index = self.sample_index_input.value()-1
            sample = self._dataset[sample_index]
            self.parameter_editor.setup_target(sample.classes, sample.distribution)
        self.parameter_editor.show()
        self.parameter_editor.update_chart()

    def on_try_fit_clicked(self):
        self.do_test()

    def on_try_previous_clicked(self):
        current = self.sample_index_input.value()
        if current == 1:
            return
        self.sample_index_input.setValue(current-1)
        self.do_test()

    def on_try_next_clicked(self):
        current = self.sample_index_input.value()
        if current == len(self._dataset):
            return
        self.sample_index_input.setValue(current+1)
        self.do_test()

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("SSU Analyzer"))
        self.control_group.setTitle(self.tr("Control"))
        self.distribution_label.setText(self.tr("Distribution Type"))
        self.component_number_label.setText(self.tr("Number of Components"))
        self.sample_index_label.setText(self.tr("Sample Index"))
        self.sample_name_label.setText(self.tr("Sample Name"))
        if self._dataset is None:
            self.sample_name_display.setText(self.tr("Unknown"))
        self.try_fit_button.setText(self.tr("Try Fit"))
        self.edit_parameter_button.setText(self.tr("Edit Parameters"))
        self.try_previous_button.setText(self.tr("Try Previous"))
        self.try_next_button.setText(self.tr("Try Next"))
        self.chart_group.setTitle(self.tr("Chart"))
        self.result_group.setTitle(self.tr("Result"))
        self.result_view.retranslate()
